# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import os
import sys
import warnings
from pathlib import Path
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import urlopen

import numpy as np

from asreview._deprecated import _deprecated_func

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points as _entry_points
else:
    from importlib_metadata import entry_points as _entry_points


def _unsafe_dict_update(default_dict, override_dict):
    """
    Using defaults and an overriding dictionary, create a new dictionary.
    This new dictionary has the same values as the default dictionary and
    the same types. Thus, if there are values that are in the overriding
    dictionary, but not in the original, they will be ignored.

    Arguments
    ---------
    default_dict: dict
        Starting dictionary with defaults.
    override_dict: dict
        Dictionary with custom values (such as model parameters).

    Returns
    -------
    dict
        Merged dictionary.
    """
    new_dict = default_dict
    for key in override_dict:
        if key not in default_dict:
            print(f"Warning: key {key} is being ignored.")

    for key in new_dict:
        if key in override_dict:
            str_val = override_dict[key]
            if isinstance(new_dict[key], bool):
                new_dict[key] = str_val in ["True", "true", "T", "t", True]
            else:
                try:
                    new_dict[key] = type(new_dict[key])(str_val)
                except TypeError:
                    raise TypeError(f"Error at {key}")
    return new_dict


def _safe_dict_update(default_dict, override_dict):
    """
    Using defaults and an overriding dictionary, create a new dictionary.
    This new dictionary has the same values as the default dictionary.
    Thus, if there are values that are in the overriding
    dictionary, but not in the original, they will be ignored.
    In contrast to the unsafe version, the type should be supplied in the
    default dictionary: key: (value, type).

    Arguments
    ---------
    default_dict: dict
        Starting dictionary with defaults.
    override_dict: dict
        Dictionary with custom values (such as model parameters).

    Returns
    -------
    dict
        Merged dictionary.
    """
    new_dict = {}
    for key in default_dict:
        new_dict[key] = default_dict[key][0]

    for key in override_dict:
        if key not in default_dict:
            print(f"Warning: key {key} is being ignored.")

    for key in new_dict:
        if key in override_dict:
            str_val = override_dict[key]
            type_val = default_dict[key][1]
            if type_val == bool:
                new_dict[key] = str_val in ["True", "true", "T", "t"]
            else:
                try:
                    new_dict[key] = type_val(str_val)
                except TypeError:
                    raise TypeError(f"Error at {key}")
    return new_dict


def _deprecated_kwarg(kwarg_map):
    def dec(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            new_kwargs = {}
            for k, v in kwargs.items():
                if k in kwarg_map:
                    warnings.warn(f"Keyword argument '{k}' is deprecated. Use '{kwarg_map[k]}' instead.", DeprecationWarning)  # noqa
                new_kwargs[kwarg_map.get(k, k)] = v
            return func(*args, **new_kwargs)
        return wrapper
    return dec


def _get_filename_from_url(url):
    if not is_url(url):
        raise ValueError(f"'{url}' is not a valid URL.")

    if Path(urlparse(url).path).suffix:
        return Path(urlparse(url).path).name
    else:
        try:
            return urlopen(url).headers.get_filename()
        except HTTPError as err:
            # 308 (Permanent Redirect) not supported
            # See https://bugs.python.org/issue40321
            if err.code == 308:
                return _get_filename_from_url(err.headers.get("Location"))
            else:
                raise err


def asreview_path():
    """Get the location where projects are stored.

    Overwrite this location by specifying the ASREVIEW_PATH enviroment
    variable.
    """
    if os.environ.get("ASREVIEW_PATH", None):
        asreview_path = Path(os.environ["ASREVIEW_PATH"])
    else:
        asreview_path = Path("~", ".asreview").expanduser()

    asreview_path.mkdir(parents=True, exist_ok=True)

    return asreview_path


def get_data_home(data_home=None):
    """Return the path of the ASR data dir.

    This folder is used by some large dataset loaders to avoid downloading the
    data several times.
    By default the data dir is set to a folder named 'asr_data' in the
    user home folder.
    Alternatively, it can be set by the 'ASR_DATA' environment
    variable or programmatically by giving an explicit folder path. The '~'
    symbol is expanded to the user home folder.
    If the folder does not already exist, it is automatically created.

    Parameters
    ----------
    data_home : str | None
        The path to scikit-learn data dir.

    """
    if data_home is None:
        data_home = os.environ.get("ASR_DATA", Path("~", "asr_data"))
    data_home = Path(data_home).expanduser()

    if not data_home.exists():
        data_home.mkdir(parents=True, exist_ok=True)

    return data_home


def format_to_str(obj):
    """Create string from object, concatenate if list."""
    if obj is None:
        return ""
    res = ""
    if isinstance(obj, list):
        res = " ".join(obj)
    else:
        res = obj
    return res


def pretty_format(result):
    longest_key = max([len(key) for key in result])
    result_str = ""
    for key, value in result.items():
        temp_str = "{{key: <{n}}}: {{value}}\n".format(n=longest_key)
        result_str += temp_str.format(key=key, value=value)
    return result_str


def is_iterable(i):
    """Check if a variable is iterable, but not a string."""
    try:
        iter(i)
        if isinstance(i, str):
            return False
        return True
    except TypeError:
        return False


@_deprecated_func(
    "list_model_names is deprecated, "
    "use asreview.models.classifiers.list_classifiers instead"
)
def list_model_names(group="asreview.models"):
    # Remove because of bug with unused default value.
    return list(_entry_points(group=group).names)


@_deprecated_kwarg({"entry_name": "group"})
def list_reader_names(group="asreview.readers"):
    return list(_entry_points(group=group).names)


@_deprecated_kwarg({"entry_name": "group"})
def list_writer_names(group="asreview.writers"):
    return list(_entry_points(group=group).names)


@_deprecated_func(
    "get_entry_points is deprecated, "
    "use _entry_points(group='asreview.entry_points') instead"
)
def get_entry_points(entry_name="asreview.entry_points"):
    """Get the entry points for asreview.

    Parameters
    ----------
    entry_name: str
        Name of the submodule. Default "asreview.entry_points".

    Returns
    -------
    dict:
        Dictionary with the name of the entry point as key
        and the entry point as value.
    """

    return {entry.name: entry for entry in _entry_points(group=entry_name)}


def is_url(url):
    """Check if object is a valid url."""
    try:
        result = urlparse(url)
        return all(
            getattr(result, x) not in [b"", ""] for x in ["scheme", "netloc", "path"])
    except Exception:
        return False


def get_random_state(random_state):
    """Create a RandomState instance.

    Parameters
    ----------
    random_state: int, numpy.RandomState
        If it is an integer, seed a new random state.
        If it is a RandomState, return it (nop).
        If it is None, return the random state of numpy.
    """

    if not isinstance(random_state, np.random.RandomState):
        return np.random.RandomState(random_state)

    return random_state


def _get_executable():
    """Get the Python executable"""

    return sys.executable if sys.executable else "python"



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import floor
from math import log

import numpy as np

from asreview.models.balance.base import BaseBalance
from asreview.models.balance.simple import SimpleBalance
from asreview.utils import get_random_state


class DoubleBalance(BaseBalance):
    """Double balance strategy (``double``).

    Class to get the two way rebalancing function and arguments.
    It super samples ones depending on the number of 0's and total number
    of samples in the training data.

    Arguments
    ---------
    a: float
        Governs the weight of the 1's. Higher values mean linearly more 1's
        in your training sample.
    alpha: float
        Governs the scaling the weight of the 1's, as a function of the
        ratio of ones to zeros. A positive value means that the lower the
        ratio of zeros to ones, the higher the weight of the ones.
    b: float
        Governs how strongly we want to sample depending on the total
        number of samples. A value of 1 means no dependence on the total
        number of samples, while lower values mean increasingly stronger
        dependence on the number of samples.
    beta: float
        Governs the scaling of the weight of the zeros depending on the
        number of samples. Higher values means that larger samples are more
        strongly penalizing zeros.
    """

    name = "double"
    label = "Dynamic resampling (Double)"

    def __init__(self, a=2.155, alpha=0.94, b=0.789, beta=1.0, random_state=None):
        super(DoubleBalance, self).__init__()
        self.a = a
        self.alpha = alpha
        self.b = b
        self.beta = beta
        self.fallback_model = SimpleBalance()
        self._random_state = get_random_state(random_state)

    def sample(self, X, y, train_idx):
        """Resample the training data.

        Arguments
        ---------
        X: numpy.ndarray
            Complete feature matrix.
        y: numpy.ndarray
            Labels for all papers.
        train_idx: numpy.ndarray
            Training indices, that is all papers that have been reviewed.

        Returns
        -------
        numpy.ndarray,numpy.ndarray:
            X_train, y_train: the resampled matrix, labels.
        """
        # Get inclusions and exclusions
        one_idx = train_idx[np.where(y[train_idx] == 1)]
        zero_idx = train_idx[np.where(y[train_idx] == 0)]

        # Fall back to simple sampling if we have only ones or zeroes.
        if len(one_idx) == 0 or len(zero_idx) == 0:
            self.fallback_model.sample(X, y, train_idx)

        n_one = len(one_idx)
        n_zero = len(zero_idx)
        n_train = n_one + n_zero

        # Compute sampling weights.
        one_weight = _one_weight(n_one, n_zero, self.a, self.alpha)
        zero_weight = _zero_weight(n_one + n_zero, self.b, self.beta)
        tot_zo_weight = one_weight * n_one + zero_weight * n_zero
        # Number of inclusions to sample.
        n_one_train = random_round(
            one_weight * n_one * n_train / tot_zo_weight, self._random_state
        )
        # Should be at least 1, and at least two spots should be for exclusions.
        n_one_train = max(1, min(n_train - 2, n_one_train))
        # Number of exclusions to sample
        n_zero_train = n_train - n_one_train

        # Sample records of ones and zeroes
        one_train_idx = fill_training(one_idx, n_one_train, self._random_state)
        zero_train_idx = fill_training(zero_idx, n_zero_train, self._random_state)
        # Merge and shuffle.
        all_idx = np.concatenate([one_train_idx, zero_train_idx])
        self._random_state.shuffle(all_idx)

        # Return resampled feature matrix and labels.
        return X[all_idx], y[all_idx]

    def full_hyper_space(self):
        from hyperopt import hp

        parameter_space = {
            "bal_a": hp.lognormal("bal_a", 0, 1),
            "bal_alpha": hp.uniform("bal_alpha", 0, 2),
            "bal_b": hp.uniform("bal_b", 0, 1),
            # "bal_beta": hp.uniform("bal_beta", 0, 2),
        }
        return parameter_space, {}


def _one_weight(n_one, n_zero, a, alpha):
    """Get the weight of the ones."""
    weight = a * (n_one / n_zero) ** (-alpha)
    return weight


def _zero_weight(n_read, b, beta):
    """Get the weight of the zeros."""
    weight = 1 - (1 - b) * (1 + log(n_read)) ** (-beta)
    return weight


def random_round(value, random_state):
    """Round up or down, depending on how far the value is.

    For example: 8.1 would be rounded to 8, 90% of the time, and rounded
    to 9, 10% of the time.
    """
    base = int(floor(value))
    if random_state.rand() < value - base:
        base += 1
    return base


def fill_training(src_idx, n_train, random_state):
    """Copy/sample until there are n_train indices sampled/copied."""
    # Number of copies needed.
    n_copy = int(n_train / len(src_idx))
    # For the remainder, use sampling.
    n_sample = n_train - n_copy * len(src_idx)

    # Copy indices
    dest_idx = np.tile(src_idx, n_copy).reshape(-1)
    # Add samples
    dest_idx = np.append(
        dest_idx, random_state.choice(src_idx, n_sample, replace=False)
    )
    return dest_idx



import json
from pathlib import Path
from shutil import copyfile
from shutil import make_archive
from zipfile import ZipFile

from asreview.project import ASReviewProject
from asreview.project import open_state
from asreview.state.legacy.json import JSONState
from asreview.state.legacy.utils import open_state as open_state_legacy
from asreview.state.legacy.utils import state_from_file
from asreview.state.sql_converter import is_converted_project
from asreview.state.sql_converter import rollback_conversion
from asreview.state.sql_converter import upgrade_asreview_project_file

OLD_STATE_FP = Path("tests", "asreview_files", "test_converter_example_old.asreview")


def compare_state_to_converted(state_fp, converted_state_fp):
    """Compare an old state file to a converted state file and
    check that the contents are the same

    Arguments
    ---------
    state_fp: path-like
        Filepath to the old project file.
    converted_state_fp: path-like
        Filepath to the converted state file.
    """
    with open(Path(state_fp.parent, "labeled.json"), "r") as file:
        labeled_json = json.load(file)
    # old_record_ids = [x[0] for x in labeled_json]
    old_labels = [x[1] for x in labeled_json]

    with open_state_legacy(state_fp) as old_state:
        old_state_length = len(old_state._state_dict["labels"])

        data_hash = list(old_state._state_dict["data_properties"].keys())[0]
        old_feature_matrix = old_state.get_feature_matrix(data_hash)
        old_settings = old_state.settings.to_dict()

    with open_state(converted_state_fp) as new_state:
        # Get data from the new state.
        new_record_ids = new_state.get_order_of_labeling().tolist()
        new_labels = new_state.get_labels().tolist()
        new_settings = new_state.settings.to_dict()

    new_project = ASReviewProject(converted_state_fp)
    feature_extraction_method = new_project.feature_matrices[0]["id"]
    new_feature_matrix = new_project.get_feature_matrix(feature_extraction_method)

    # Compare data.
    # assert old_indices == new_record_ids
    assert max(new_record_ids) < old_state_length
    assert old_labels == new_labels
    # assert old_query_strategies == new_query_strategies
    assert (old_feature_matrix != new_feature_matrix).nnz == 0
    assert old_settings == new_settings


def test_converter(tmpdir):
    # Copy old project file to temporary folder.
    converted_fp = Path(tmpdir, "converted.asreview")
    copyfile(OLD_STATE_FP, converted_fp)

    # Unzip the converted state file.
    unzipped_fp = Path(tmpdir, "unzipped.asreview")
    with ZipFile(converted_fp) as zipobj:
        zipobj.extractall(unzipped_fp)
    converted_fp = unzipped_fp
    # -------------------------------------------------
    # Convert the old project file to a new state file.
    upgrade_asreview_project_file(converted_fp, from_version=0, to_version=1)

    # Check that the contents are the same.
    compare_state_to_converted(
        Path(converted_fp, "legacy", "result.json"), converted_fp
    )

    # Check if the rollback works.
    assert is_converted_project(converted_fp)
    rollback_conversion(converted_fp)

    # Zip the converted file to allow for reading it.
    zipped_fp = make_archive(Path(tmpdir, "zipped"), "zip", converted_fp)
    zipped_fp = Path(zipped_fp).rename(Path(tmpdir, "zipped.asreview"))

    state = state_from_file(zipped_fp)[zipped_fp.name]
    assert isinstance(state, JSONState)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def convert_id_to_idx(data_obj, record_id):
    """Convert record_id to row number."""

    inv_record_id = dict(zip(data_obj.df.index.tolist(), range(len(data_obj))))

    result = []
    for i in record_id:
        try:
            result.append(inv_record_id[i])
        except KeyError:
            raise KeyError(f"record_id {i} not found in data.")

    return result



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import shutil
import tempfile
import time
import zipfile
from contextlib import contextmanager
from datetime import datetime
from functools import wraps
from pathlib import Path
from uuid import uuid4

import jsonschema
import numpy as np
from filelock import FileLock
from scipy.sparse import csr_matrix
from scipy.sparse import load_npz
from scipy.sparse import save_npz

from asreview._version import get_versions
from asreview.config import LABEL_NA
from asreview.config import PROJECT_MODE_SIMULATE
from asreview.config import PROJECT_MODES
from asreview.config import SCHEMA
from asreview.data import ASReviewData
from asreview.state.errors import StateNotFoundError
from asreview.state.sqlstate import SQLiteState
from asreview.utils import asreview_path

PATH_PROJECT_CONFIG = "project.json"
PATH_PROJECT_CONFIG_LOCK = "project.json.lock"
PATH_FEATURE_MATRICES = "feature_matrices"


class ProjectError(Exception):
    pass


class ProjectExistsError(Exception):
    pass


class ProjectNotFoundError(Exception):
    pass


def get_project_path(folder_id):
    """Get the project directory.

    Arguments
    ---------
    folder_id: str
        The id of the folder containing a project. If there is no
        authentication, the folder_id is equal to the project_id. Otherwise,
        this is equal to {project_owner_id}_{project_id}.
    """
    return Path(asreview_path(), folder_id)


def project_from_id(f):
    """Decorator function that takes a user account as parameter,
    the user account is used to get the correct sub folder in which
    the projects is
    """

    @wraps(f)
    def decorated_function(project_id, *args, **kwargs):
        project_path = get_project_path(project_id)
        if not is_project(project_path):
            raise ProjectNotFoundError(f"Project '{project_id}' not found")
        project = ASReviewProject(project_path, project_id=project_id)
        return f(project, *args, **kwargs)

    return decorated_function


def get_projects(project_paths=None):
    """Get the ASReview projects at the given paths.

    Arguments
    ---------
    project_paths : list[Path], optional
        List of paths to projects. By default all the projects in the asreview
        folder are used, by default None

    Returns
    -------
    list[ASReviewProject]
        Projects at the given project paths.
    """
    if project_paths is None:
        project_paths = [path for path in asreview_path().iterdir() if path.is_dir()]

    return [ASReviewProject(project_path) for project_path in project_paths]


def is_project(project_path):
    project_path = Path(project_path) / PATH_PROJECT_CONFIG

    return project_path.exists()


def is_v0_project(project_path):
    """Check if a project file is of a ASReview version 0 project."""

    return not Path(project_path, "reviews").exists()


@contextmanager
def open_state(asreview_obj, review_id=None, read_only=True):
    """Initialize a state class instance from a project folder.

    Arguments
    ---------
    asreview_obj: str/pathlike/ASReviewProject
        Filepath to the (unzipped) project folder or ASReviewProject object.
    review_id: str
        Identifier of the review from which the state will be instantiated.
        If none is given, the first review in the reviews folder will be taken.
    read_only: bool
        Whether to open in read_only mode.

    Returns
    -------
    SQLiteState
    """

    # Unzip the ASReview data if needed.
    if isinstance(asreview_obj, ASReviewProject):
        project = asreview_obj
    elif zipfile.is_zipfile(asreview_obj) and Path(asreview_obj).suffix == ".asreview":
        if not read_only:
            raise ValueError("ASReview files do not support not read only files.")

        # work from a temp dir
        tmpdir = tempfile.TemporaryDirectory()
        project = ASReviewProject.load(asreview_obj, tmpdir.name)
    else:
        project = ASReviewProject(asreview_obj)

    # init state class
    state = SQLiteState(read_only=read_only)

    try:
        if len(project.reviews) > 0:
            if review_id is None:
                review_id = project.config["reviews"][0]["id"]
            logging.debug(f"Opening review {review_id}.")
            state._restore(project.project_path, review_id)
        elif len(project.reviews) == 0 and not read_only:
            review_id = uuid4().hex
            logging.debug(f"Create new review (state) with id {review_id}.")
            state._create_new_state_file(project.project_path, review_id)
            project.add_review(review_id)
        else:
            raise StateNotFoundError(
                "State file does not exist, and in " "read only mode."
            )
        yield state
    finally:
        try:
            state.close()
        except AttributeError:
            # file seems to be closed, do nothing
            pass


class ASReviewProject:
    """Project class for ASReview project files."""

    def __init__(self, project_path, project_id=None):
        self.project_path = Path(project_path)
        self.project_id = project_id

    @classmethod
    def create(
        cls,
        project_path,
        project_id=None,
        project_mode="oracle",
        project_name=None,
        project_description=None,
        project_authors=None,
    ):
        """Initialize the necessary files specific to the web app."""

        project_path = Path(project_path)

        if is_project(project_path):
            raise ProjectExistsError("Project already exists.")

        if project_mode not in PROJECT_MODES:
            raise ValueError(
                f"Project mode '{project_mode}' is not in " f"{PROJECT_MODES}."
            )

        if project_id is None:
            project_id = project_path.stem

        if project_name is None:
            project_name = project_path.stem

        if project_path.is_dir():
            raise IsADirectoryError(f"Project folder {project_path} already exists.")

        try:
            project_path.mkdir(parents=True, exist_ok=True)
            Path(project_path, "data").mkdir(exist_ok=True)
            Path(project_path, PATH_FEATURE_MATRICES).mkdir(exist_ok=True)
            Path(project_path, "reviews").mkdir(exist_ok=True)

            config = {
                "version": get_versions()["version"],
                "id": project_id,
                "mode": project_mode,
                "name": project_name,
                "description": project_description,
                "authors": project_authors,
                "created_at_unix": int(time.time()),
                "datetimeCreated": str(datetime.now()),
                "reviews": [],
                "feature_matrices": [],
            }

            # validate new config before storing
            jsonschema.validate(instance=config, schema=SCHEMA)

            project_fp = Path(project_path, PATH_PROJECT_CONFIG)
            project_fp_lock = Path(project_path, PATH_PROJECT_CONFIG_LOCK)
            lock = FileLock(project_fp_lock, timeout=3)

            # create a file with project info
            with lock:
                with open(project_fp, "w") as f:
                    json.dump(config, f)

        except Exception as err:
            # remove all generated folders and raise error
            shutil.rmtree(project_path)
            raise err

        return cls(project_path, project_id=project_id)

    @property
    def config(self):
        try:
            return self._config
        except AttributeError:
            project_fp = Path(self.project_path, PATH_PROJECT_CONFIG)
            project_fp_lock = Path(self.project_path, PATH_PROJECT_CONFIG_LOCK)
            lock = FileLock(project_fp_lock, timeout=3)

            try:
                with lock:
                    # read the file with project info
                    with open(project_fp, "r") as fp:
                        config = json.load(fp)
                        self._config = config

                        return config

            except FileNotFoundError:
                raise ProjectNotFoundError(f"Project '{self.project_path}' not found")

    @config.setter
    def config(self, config):
        project_fp = Path(self.project_path, PATH_PROJECT_CONFIG)
        project_fp_lock = Path(self.project_path, PATH_PROJECT_CONFIG_LOCK)
        lock = FileLock(project_fp_lock, timeout=3)

        with lock:
            with open(project_fp, "w") as f:
                json.dump(config, f)

        self._config = config

    def update_config(self, **kwargs):
        """Update project info"""

        kwargs_copy = kwargs.copy()

        # validate schema
        if "mode" in kwargs_copy and kwargs_copy["mode"] not in PROJECT_MODES:
            raise ValueError("Project mode '{}' not found.".format(kwargs_copy["mode"]))

        # update project file
        config = self.config
        config.update(kwargs_copy)

        # validate new config before storing
        jsonschema.validate(instance=config, schema=SCHEMA)

        self.config = config
        return config

    def add_dataset(self, file_name):
        """Add file path to the project file.

        Add file to data subfolder and fill the pool of iteration 0.
        """
        self.update_config(dataset_path=file_name)

        # fill the pool of the first iteration
        fp_data = Path(self.project_path, "data", self.config["dataset_path"])
        as_data = ASReviewData.from_file(fp_data)

        with open_state(self.project_path, read_only=False) as state:
            # save the record ids in the state file
            state.add_record_table(as_data.record_ids)

            # if the data contains labels, add them to the state file
            if (
                self.config["mode"] != PROJECT_MODE_SIMULATE
                and as_data.labels is not None
            ):
                labeled_indices = np.where(as_data.labels != LABEL_NA)[0]
                labels = as_data.labels[labeled_indices].tolist()
                labeled_record_ids = as_data.record_ids[labeled_indices].tolist()

                # add the labels as prior data
                state.add_labeling_data(
                    record_ids=labeled_record_ids,
                    labels=labels,
                    notes=[None for _ in labeled_record_ids],
                    prior=True,
                )

    def remove_dataset(self):
        """Remove dataset from project."""
        # reset dataset_path
        self.update_config(dataset_path=None)

        # remove datasets from project
        shutil.rmtree(Path(self.project_path, "data"))

        # remove state file if present
        if Path(self.project_path, "reviews").is_dir() and any(
            Path(self.project_path, "reviews").iterdir()
        ):
            self.delete_review()

    def clean_tmp_files(self):
        """Clean temporary files in a project.

        Arguments
        ---------
        project_id: str
            The id of the current project.
        """

        # clean pickle files
        for f_pickle in self.project_path.rglob("*.pickle"):
            try:
                os.remove(f_pickle)
            except OSError as e:
                print(f"Error: {f_pickle} : {e.strerror}")

    @property
    def feature_matrices(self):
        try:
            return self.config["feature_matrices"]
        except Exception:
            return []

    def add_feature_matrix(self, feature_matrix, feature_extraction_method):
        """Add feature matrix to project file.

        Arguments
        ---------
        feature_matrix: numpy.ndarray, scipy.sparse.csr.csr_matrix
            The feature matrix to add to the project file.
        feature_extraction_method: str
            Name of the feature extraction method.
        """
        # Make sure the feature matrix is in csr format.
        if isinstance(feature_matrix, np.ndarray):
            feature_matrix = csr_matrix(feature_matrix)
        if not isinstance(feature_matrix, csr_matrix):
            raise ValueError(
                "The feature matrix should be convertible to type "
                "scipy.sparse.csr.csr_matrix."
            )

        matrix_filename = f"{feature_extraction_method}_feature_matrix.npz"
        save_npz(
            Path(self.project_path, PATH_FEATURE_MATRICES, matrix_filename),
            feature_matrix,
        )

        # Add the feature matrix to the project config.
        config = self.config

        feature_matrix_config = {
            "id": feature_extraction_method,
            "filename": matrix_filename,
        }

        # Add container for feature matrices.
        if "feature_matrices" not in config:
            config["feature_matrices"] = []

        config["feature_matrices"].append(feature_matrix_config)

        self.config = config

    def get_feature_matrix(self, feature_extraction_method):
        """Get the feature matrix from the project file.

        Arguments
        ---------
        feature_extraction_method: str
            Name of the feature extraction method for which to get the matrix.

        Returns
        -------
        scipy.sparse.csr_matrix:
            Feature matrix in sparse format.
        """
        matrix_filename = f"{feature_extraction_method}_feature_matrix.npz"
        return load_npz(Path(self.project_path, PATH_FEATURE_MATRICES, matrix_filename))

    @property
    def reviews(self):
        try:
            return self.config["reviews"]
        except Exception:
            return []

    def add_review(self, review_id, start_time=None, status="setup"):
        """Add new review metadata.

        Arguments
        ---------
        review_id: str
            The review_id uuid4.
        status: str
            The status of the review. One of 'setup', 'running',
            'finished'.
        start_time:
            Start of the review.

        """
        if start_time is None:
            start_time = datetime.now()

        # Add the review to the project.
        config = self.config

        review_config = {
            "id": review_id,
            "start_time": str(start_time),
            "status": status
            # "end_time": datetime.now()
        }

        # add container for reviews
        if "reviews" not in config:
            config["reviews"] = []

        config["reviews"].append(review_config)

        self.config = config

    def update_review(self, review_id=None, **kwargs):
        """Update review metadata.

        Arguments
        ---------
        review_id: str
            The review_id uuid4. Default None, which is the
            first added review.
        status: str
            The status of the review. One of 'setup', 'running',
            'finished'.
        start_time:
            Start of the review.
        end_time: End time of the review.
        """

        # read the file with project info
        config = self.config

        if review_id is None:
            review_index = 0
        else:
            review_index = [x["id"] for x in self.config["reviews"]].index(review_id)

        review_config = config["reviews"][review_index]
        review_config.update(kwargs)

        config["reviews"][review_index] = review_config

        # update the file with project info
        self.config = config

    def delete_review(self, remove_folders=False):
        try:
            # remove the folder tree
            shutil.rmtree(Path(self.project_path, PATH_FEATURE_MATRICES))

            # recreate folder structure if True
            if not remove_folders:
                Path(self.project_path, PATH_FEATURE_MATRICES).mkdir(exist_ok=True)
        except Exception:
            print("Failed to remove feature matrices.")

        try:
            path_review = Path(self.project_path, "reviews")
            shutil.rmtree(path_review)
            if not remove_folders:
                Path(self.project_path, "reviews").mkdir(exist_ok=True)
        except Exception:
            print("Failed to remove sql database.")

        # update the config
        self.update_config(**{"reviews": [], "feature_matrices": []})

    def mark_review_finished(self, review_id=None):
        """Mark a review in the project as finished.

        If no review_id is given, mark the first review as finished.

        Arguments
        ---------
        review_id: str
            Identifier of the review to mark as finished.
        """

        self.update_review(
            review_id=review_id, status="finished", end_time=str(datetime.now())
        )

    def export(self, export_fp):
        if Path(export_fp).suffix != ".asreview":
            raise ValueError("Export file should have .asreview extension.")

        if Path(export_fp) == Path(self.project_path):
            raise ValueError("export_fp should not be identical to project path.")

        export_fp_tmp = Path(export_fp).with_suffix(".asreview.zip")

        # copy the source tree, but ignore pickle files
        shutil.copytree(
            self.project_path,
            export_fp_tmp,
            ignore=shutil.ignore_patterns("*.pickle", "*.lock"),
        )

        # create the archive
        shutil.make_archive(export_fp_tmp, "zip", root_dir=export_fp_tmp)

        # remove the unzipped folder and move zip
        shutil.rmtree(export_fp_tmp)
        shutil.move(f"{export_fp_tmp}.zip", export_fp)

    @classmethod
    def load(cls, asreview_file, project_path, safe_import=False):
        tmpdir = tempfile.TemporaryDirectory().name

        try:
            # Unzip the project file
            with zipfile.ZipFile(asreview_file, "r") as zip_obj:
                zip_filenames = zip_obj.namelist()

                # raise error if no ASReview project file
                if PATH_PROJECT_CONFIG not in zip_filenames:
                    raise ValueError("Project file is not valid project.")

                # extract all files to folder
                for f in zip_filenames:
                    if not f.endswith(".pickle"):
                        zip_obj.extract(f, path=tmpdir)

        except zipfile.BadZipFile:
            raise ValueError("File is not an ASReview file.")

        with open(Path(tmpdir, PATH_PROJECT_CONFIG), "r") as f:
            project_config = json.load(f)

        if safe_import:
            # assign a new id to the project.
            project_config["id"] = uuid4().hex
            with open(Path(tmpdir, PATH_PROJECT_CONFIG), "r+") as f:
                # write to file
                f.seek(0)
                json.dump(project_config, f)
                f.truncate()

        # location to copy file to
        # Move the project from the temp folder to the projects folder.
        os.replace(tmpdir, Path(project_path, project_config["id"]))

        return cls(Path(project_path, project_config["id"]))

    def set_error(self, err, save_error_message=True):
        err_type = type(err).__name__
        self.update_review(status="error")

        # write error to file if label method is prior (first iteration)
        if save_error_message:
            message = {
                "message": f"{err_type}: {err}",
                "type": f"{err_type}",
                "datetime": str(datetime.now()),
            }

            with open(Path(self.project_path, "error.json"), "w") as f:
                json.dump(message, f)

    def remove_error(self, status):
        error_path = self.project_path / "error.json"
        if error_path.exists():
            try:
                os.remove(error_path)
            except Exception as err:
                raise ValueError(f"Failed to clear the error. {err}")
        self.update_review(status=status)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import rispy


class RISWriter:
    """RIS file writer."""

    name = "ris"
    label = "RIS"
    caution = "Available only if you imported a RIS file when creating the project"
    write_format = ".ris"

    @classmethod
    def write_data(cls, df, fp, labels=None, ranking=None):
        """Export dataset.

        Arguments
        ---------
        df: pandas.Dataframe
            Dataframe of all available record data.
        fp: str, pathlib.Path
            File path to the RIS file, if exists.
        labels: list, numpy.ndarray
            Current labels will be overwritten by these labels
            (including unlabelled). No effect if labels is None.
        ranking: list
            Reorder the dataframe according to these (internal) indices.
            Default ordering if ranking is None.

        Returns
        -------
        RIS file
            Dataframe of all available record data.
        """

        # Turn pandas DataFrame into records (list of dictionaries) for rispy
        records = df.to_dict("records")

        # Create an array for storing modified records
        records_new = []

        # Iterate over all available records
        for rec in records:

            def _notnull(v):
                if isinstance(v, list):
                    return False
                return pd.notnull(v)

            # Remove all nan values
            rec_copy = {k: v for k, v in rec.items() if _notnull(v)}

            for m in ["authors", "keywords", "notes"]:  # AU, KW, N1
                # Check the "authors" - AU
                try:
                    rec_copy[m] = eval(rec_copy[m])
                except Exception:
                    rec_copy[m] = []

            # Get label for record if specified, if not specified set to -1
            included = rec_copy.pop("included", -1)

            # Map labels to notes
            dict_note = {
                -1: "ASReview_not_seen",
                0: "ASReview_irrelevant",
                1: "ASReview_relevant",
            }
            rec_copy["notes"].insert(0, dict_note[included])

            # Append the deepcopied and updated record to a new array
            records_new.append(rec_copy)

        # From buffered dataframe
        if fp is None:
            # Write the whole content to buffer
            return rispy.dumps(records_new)

        # From IO dataframe
        else:
            # Write the whole content to a file
            with open(fp, "w", encoding="utf8") as fp:
                rispy.dump(records_new, fp)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from asreview.io.csv_reader import CSVReader
from asreview.io.csv_writer import CSVWriter
from asreview.io.excel_reader import ExcelReader
from asreview.io.excel_writer import ExcelWriter
from asreview.io.paper_record import PaperRecord
from asreview.io.ris_reader import RISReader
from asreview.io.ris_writer import RISWriter
from asreview.io.tsv_writer import TSVWriter
from asreview.io.utils import list_readers
from asreview.io.utils import list_writers



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

import pytest
from sqlalchemy.orm import close_all_sessions

from asreview.webapp import DB
from asreview.webapp.start_flask import create_app
from asreview.webapp.tests.utils import crud

PROJECTS = [
    {
        "mode": "explore",
        "name": "demo project",
        "authors": "asreview team",
        "description": "hello world",
    },
    {
        "mode": "explore",
        "name": "another demo project",
        "authors": "asreview team",
        "description": "hello world",
    },
]


def _get_app(app_type="auth-basic", path=None):
    """Create and returns test flask app based on app_type"""
    # set asreview path
    os.environ.update({"ASREVIEW_PATH": path})
    # get path of appropriate flask config
    base_dir = Path(__file__).resolve().parent / "config"
    if app_type == "auth-basic":
        config_path = str(base_dir / "auth_basic_config.toml")
    elif app_type == "auth-no-creation":
        config_path = str(base_dir / "auth_no_creation.toml")
    elif app_type == "auth-verified":
        config_path = str(base_dir / "auth_verified_config.toml")
    elif app_type == "no-auth":
        config_path = str(base_dir / "no_auth_config.toml")
    else:
        raise ValueError(f"Unknown config {app_type}")
    # create app
    app = create_app(flask_configfile=config_path)
    # and return it
    return app


@pytest.fixture(scope="function", autouse=True)
def asreview_path_fixture(tmp_path_factory):
    """Fixture that creates and removes the ASReview test
    directory for the entire session."""
    # create an ASReview folder
    asreview_path = tmp_path_factory.mktemp("asreview-test")
    assert Path(asreview_path).exists()
    assert len(list(Path(asreview_path).glob('*'))) == 0
    yield str(asreview_path.absolute())
    # Pytest handles removal of ASReview folder


# unauthenticated app
@pytest.fixture
def unauth_app(asreview_path_fixture):
    """Create an unauthenticated version of the app."""
    # create the app
    app = _get_app("no-auth", path=asreview_path_fixture)
    with app.app_context():
        yield app


# authenticated app
@pytest.fixture
def auth_app(asreview_path_fixture):
    """Create an authenticated app, account creation
    allowed."""
    # create app
    app = _get_app(path=asreview_path_fixture)
    with app.app_context():
        yield app


@pytest.fixture
def client_auth(asreview_path_fixture):
    """Flask client for basic authenticated app, account
    creation allowed."""
    app = _get_app("auth-basic", path=asreview_path_fixture)
    with app.app_context():
        yield app.test_client()
        crud.delete_everything(DB)
        close_all_sessions()
        DB.engine.raw_connection().close()


@pytest.fixture
def client_auth_no_creation(asreview_path_fixture):
    """Flask client for an authenticated app, account
    creation not allowed."""
    app = _get_app("auth-no-creation", path=asreview_path_fixture)
    with app.app_context():
        yield app.test_client()
        crud.delete_everything(DB)
        close_all_sessions()
        DB.engine.raw_connection().close()


@pytest.fixture
def client_auth_verified(asreview_path_fixture):
    """Flask client for an authenticated app, account
    creation allowed, user accounts needs account
    verification."""
    app = _get_app("auth-verified", path=asreview_path_fixture)
    with app.app_context():
        yield app.test_client()
        crud.delete_everything(DB)
        close_all_sessions()
        DB.engine.raw_connection().close()


@pytest.fixture
def client_no_auth(asreview_path_fixture):
    """Flask client for an unauthenticated app."""
    app = _get_app("no-auth", path=asreview_path_fixture)
    # make sure we have the asreview_path
    with app.app_context():
        yield app.test_client()



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.entry_points.base import BaseEntryPoint
from asreview.models.balance import list_balance_strategies
from asreview.models.classifiers import list_classifiers
from asreview.models.feature_extraction import list_feature_extraction
from asreview.models.query import list_query_strategies


def _format_algorithm(values, name, description):
    s = f"  {name: <20}Available {description}:\n\n"

    result = []

    for x in values:
        if hasattr(x, "label"):
            result.append(
                " " * 22 + f"{x.name}" + " " * (16 - len(x.name)) + f"{x.label}"
            )
        else:
            result.append(" " * 22 + f"{x.name}")

    s += "\n".join(result)
    s += "\n\n"

    return s


class AlgorithmsEntryPoint(BaseEntryPoint):
    """Entry point to list available algorithms in ASReview LAB."""

    description = "Available active learning algorithms for ASReview."

    def execute(self, argv):
        s = "Available active learning algorithms for ASReview. \n\n"

        # feature_extraction
        s += _format_algorithm(
            values=list_feature_extraction(),
            name="feature_extraction",
            description="feature extraction algorithms",
        )

        # classifiers
        s += _format_algorithm(
            values=list_classifiers(),
            name="classifiers",
            description="classification algorithms",
        )

        # query_strategies
        s += _format_algorithm(
            values=list_query_strategies(),
            name="query_strategies",
            description="query strategies",
        )

        # balance_strategies
        s += _format_algorithm(
            values=list_balance_strategies(),
            name="balance_strategies",
            description="balance strategies",
        )

        print(s)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def type_n_queries(value):
    """Custom type used for --n_queries argument.

    Parameters
    ----------
    value: str
        The argument value for --n_queries

    Returns
    -------
    type_n_queries:
        A string containing 'min' or an integer.
    """
    if value == "min":
        return value
    else:
        try:
            value_i = int(value)

            # convert -1 to None
            if value_i == -1:
                return None

            return value_i
        except ValueError:
            raise ValueError("Expected 'min' or a valid integer")



import datetime as dt
from inspect import getfullargspec

import pytest

import asreview.webapp.tests.utils.api_utils as au
import asreview.webapp.tests.utils.crud as crud
from asreview.webapp import DB
from asreview.webapp.tests.utils.config_parser import get_user

# ###################
# SIGNUP
# ###################


# test that creating a user when the app runs a no-creation
# policy, is impossible
def test_impossible_to_signup_when_not_allowed(client_auth_no_creation):
    # get user data
    user = get_user(1)
    # post form data
    status_code, data = au.signup_user(client_auth_no_creation, user)
    # check if we get a 400 status
    assert status_code == 400
    assert data["message"] == "The app is not configured to create accounts"


# Successful signup returns a 200 but with an unconfirmed user and
# an email token
def test_successful_signup_confirmed(client_auth_verified):
    # get user data
    user = get_user(1)
    # post form data
    status_code, data = au.signup_user(client_auth_verified, user)
    # check if we get a 200 status
    assert status_code == 201
    assert (
        data["message"]
        == f"An email has been sent to {user.email} "
        + "to verify your account. Please follow instructions."
    )


# test basic signing up
def test_successful_signup_no_confirmation(client_auth):
    # get user data
    user = get_user(1)
    # post form data
    status_code, data = au.signup_user(client_auth, user)
    # check if we get a 201 status
    assert status_code == 201
    assert data["message"] == f'User "{user.email}" created.'


# Adding an existing identifier must return a 404 status and
# appropriate message
def test_unique_identifier(client_auth):
    # get user data
    user = get_user(1)
    # insert this user
    crud.create_user(DB, user)
    # try to create the same user again with the api
    status_code, data = au.signup_user(client_auth, user)
    assert status_code == 403
    assert data["message"] == f'User with email "{user.email}" already exists.'


# Adding an existing email must return a 404 status and
# appropriate message
def test_unique_email(client_auth):
    # get user data
    user1 = get_user(1)
    user2 = get_user(2)
    # insert user1
    crud.create_user(DB, user1)
    # try to create the user2 with the same email as user1 with the api
    user2.email = user1.email
    status_code, data = au.signup_user(client_auth, user2)
    assert status_code == 403
    assert data["message"] == f'User with email "{user2.email}" already exists.'


# ###################
# SIGNIN
# ###################


# Verified user creation: user can not signin with unconfirmed account
def test_unsuccessful_signin_with_unconfirmed_account(client_auth_verified):
    # get user data
    user = get_user(1)
    # create user with signup
    status_code, data = au.signup_user(client_auth_verified, user)
    # check if we get a 201 status
    assert status_code == 201
    # try to sign in
    status_code, data = au.signin_user(client_auth_verified, user)
    assert status_code == 404
    assert data["message"] == f"User account {user.email} is not confirmed."


# Successfully signing in a user must return a 200 response
def test_successful_signin(client_auth):
    # get user data
    user = get_user(1)
    # create user with signup, no confirmation
    status_code, data = au.signup_user(client_auth, user)
    # check if we get a 201 status
    assert status_code == 201
    # signin
    status_code, data = au.signin_user(client_auth, user)
    assert status_code == 200
    assert data["message"] == f"User {user.identifier} is logged in."


# Wrong password must return a 404 response with and an appropriate response
def test_unsuccessful_signin_wrong_password(client_auth):
    # get user data
    user = get_user(1)
    # create user with signup, no confirmation
    status_code, data = au.signup_user(client_auth, user)
    # check if we get a 201 status
    assert status_code == 201
    # change password
    user.password = "wrong_password"
    # signin
    status_code, data = au.signin_user(client_auth, user)
    assert status_code == 404
    assert data["message"] == f"Incorrect password for user {user.identifier}."


# Wrong email must return a 404 response and with an appropriate response
def test_unsuccessful_signin_wrong_email(client_auth):
    # get user data
    user = get_user(1)
    # create user with signup, no confirmation
    status_code, data = au.signup_user(client_auth, user)
    # check if we get a 201 status
    assert status_code == 201
    # change email and identifier
    user.email = "wrong_email@asreview.nl"
    user.identifier = "wrong_email@asreview.nl"
    # signin
    status_code, data = au.signin_user(client_auth, user)
    assert status_code == 404
    assert data["message"] == f"User account {user.identifier} does not exist."


# ###################
# SIGNOUT
# ###################


# Signing out must return a 200 status and an appropriate message
def test_signout(client_auth):
    # create user
    user = au.create_and_signin_user(client_auth)
    # signout
    status_code, data = au.signout_user(client_auth)
    # expect a 200
    assert status_code == 200
    assert (
        data["message"]
        == f"User with identifier {user.identifier} has been signed out."
    )


# ###################
# CONFIRMATION
# ###################


# A new token is created on signup, that token is can be confirmed
# by the confirm route
def test_token_confirmation_after_signup(client_auth_verified):
    # signup user
    user = get_user(1)
    status_code, data = au.signup_user(client_auth_verified, user)
    # refresh user
    user = crud.get_user_by_identifier(user.identifier)
    # now we confirm this user
    status_code, data = au.confirm_user(client_auth_verified, user)
    assert status_code == 200
    assert data["message"] == f"User {user.identifier} confirmed."


# A token expires in 24 hours, test confirmation response after
# 24 hours
def test_expired_token(client_auth_verified):
    # signup user
    user = get_user(1)
    status_code, data = au.signup_user(client_auth_verified, user)
    # refresh user
    user = crud.get_user_by_identifier(user.identifier)
    # manipulate token_created_at
    new_created_at = user.token_created_at - dt.timedelta(hours=28)
    user.token_created_at = new_created_at
    DB.session.commit()
    # now we try to confirm this user
    status_code, data = au.confirm_user(client_auth_verified, user)
    assert status_code == 403
    assert "token has expired" in data["message"]


# Confirmation user: if the user can't be found, this route should
# return a 404
def test_if_this_route_returns_404_user_not_found(client_auth_verified):
    # signup user
    user = get_user(1)
    status_code, data = au.signup_user(client_auth_verified, user)
    # make sure the user account is created
    assert crud.count_users() == 1
    # we keep the user model as is, not retrieving it from the DB
    # which ensures an id-less object that can be manipulated
    user.id = 100
    # now we try to confirm this user
    status_code, data = au.confirm_user(client_auth_verified, user)
    assert status_code == 404
    assert data["message"] == "No user account / correct token found."


# If the token cant be found, this route should return a 404
def test_if_this_route_returns_404_token_not_found(client_auth_verified):
    # signup user
    user = get_user(1)
    status_code, data = au.signup_user(client_auth_verified, user)
    # make sure the user account is created
    assert crud.count_users() == 1
    # we keep the user model as is, not retrieving it from the DB
    # which ensures an id-less object that can be manipulated
    user.token = "wrong_token"
    # now we try to confirm this user
    status_code, data = au.confirm_user(client_auth_verified, user)
    assert status_code == 404
    assert data["message"] == "No user account / correct token found."


# If we are not doing verification this route should return a 400
def test_confirm_route_returns_400_if_app_not_verified(client_auth):
    # signup user
    user = get_user(1)
    status_code, data = au.signup_user(client_auth, user)
    # refresh user object
    user = crud.get_user_by_identifier(user.identifier)
    # now we try to confirm this user
    status_code, data = au.confirm_user(client_auth, user)
    assert status_code == 400
    assert data["message"] == "The app is not configured to verify accounts."


# ###################
# PROFILE
# ###################


# Test user data if we request is
@pytest.mark.parametrize(
    "attribute", ["email", "identifier", "name", "origin", "affiliation"]
)
def test_get_profile(client_auth, attribute):
    user = au.create_and_signin_user(client_auth)
    # get profile
    status_code, data = au.get_profile(client_auth)
    assert status_code == 200
    # assert if none is blank
    assert data["message"][attribute] != ""
    # compare with user
    assert data["message"][attribute] == getattr(user, attribute)


# Test profile data not returned when user id does not exists
def test_get_profile_if_user_id_does_not_exist(client_auth):
    au.create_and_signin_user(client_auth)
    # remove this user from the database
    crud.delete_users(DB)
    # get profile
    status_code, data = au.get_profile(client_auth)
    assert status_code == 404
    assert data["message"] == "No user found."


# #####################
# FORGOT/RESET PASSWORD
# #####################


# Test forgot_password can't be called if we don't run an
# email config
def test_forgot_password_no_email_config(client_auth):
    # signup user
    user = get_user(1)
    status_code, data = au.signup_user(client_auth, user)
    # forgot password
    status_code, data = au.forgot_password(client_auth, user)
    assert status_code == 404
    assert data["message"] == "Forgot-password feature is not used in this app."


# Test forgot_password works under normal circumstances
def test_forgot_password_works(client_auth_verified):
    # signup user
    user = get_user(1)
    status_code, data = au.signup_user(client_auth_verified, user)
    # forgot password
    status_code, data = au.forgot_password(client_auth_verified, user)
    assert status_code == 200
    assert data["message"] == f"An email has been sent to {user.email}"


# Test forgot_password: user not found
def test_forgot_password_no_user(client_auth_verified):
    # get user, no signup
    user = get_user(1)
    # forgot password
    status_code, data = au.forgot_password(client_auth_verified, user)
    assert status_code == 404
    assert data["message"] == f'User with email "{user.email}" not found.'


# Test forgot_password: origin is not "asreview"
def test_forgot_password_wrong_origin(client_auth_verified):
    # signup user
    user = get_user(1)
    status_code, data = au.signup_user(client_auth_verified, user)
    # get fresh user object and change origin
    user = crud.update_user(DB, user, "origin", "github")
    # forgot password
    status_code, data = au.forgot_password(client_auth_verified, user)
    assert status_code == 404
    assert data["message"] == f"Your account has been created with {user.origin}."


# Test resetting password when not configured
def test_reset_password_no_email_config(client_auth):
    # signup user
    user = get_user(1)
    status_code, data = au.signup_user(client_auth, user)
    # get user
    user = crud.get_user_by_identifier(user.identifier)
    user.password = "NewPassword123!"
    # forgot password
    status_code, data = au.reset_password(client_auth, user)
    assert status_code == 404
    assert data["message"] == "Reset-password feature is not used in this app."


# Test resetting password
def test_reset_password(client_auth_verified):
    # signup user
    user = get_user(1)
    au.signup_user(client_auth_verified, user)
    # forgot password
    au.forgot_password(client_auth_verified, user)
    # get user and provide new password
    user = crud.get_user_by_identifier(user.identifier)
    user.password = "NewPassword123!"
    # reset it
    status_code, data = au.reset_password(client_auth_verified, user)
    assert status_code == 200
    assert data["message"] == "Password updated."


# Test reset password: id not found
def test_reset_password_with_wrong_user_id(client_auth_verified):
    # signup user
    user = get_user(1)
    au.signup_user(client_auth_verified, user)
    # forgot password
    au.forgot_password(client_auth_verified, user)
    # get user and provide new password
    user = crud.get_user_by_identifier(user.identifier)
    user.password = "NewPassword123!"
    # and remove from database to manipulate user-not-found
    crud.delete_users(DB)
    # reset it
    status_code, data = au.reset_password(client_auth_verified, user)
    assert status_code == 404
    assert (
        data["message"]
        == "User not found, try restarting the forgot-password procedure."
    )


# Test reset password: token is stale
def test_reset_password_with_stale_token(client_auth_verified):
    # signup user
    user = get_user(1)
    au.signup_user(client_auth_verified, user)
    # forgot password
    au.forgot_password(client_auth_verified, user)
    # get user and provide new password
    user = crud.get_user_by_identifier(user.identifier)
    user.password = "NewPassword123!"
    new_created_at = user.token_created_at - dt.timedelta(hours=28)
    user.token_created_at = new_created_at
    DB.session.commit()
    # reset password
    status_code, data = au.reset_password(client_auth_verified, user)
    assert status_code == 404
    assert (
        data["message"]
        == "Token is invalid or too old, restart the forgot-password procedure."
    )

    # Test reset password: invalid password
    # signup user
    user = get_user(1)
    au.signup_user(client_auth_verified, user)
    # forgot password
    au.forgot_password(client_auth_verified, user)
    # get user and provide new password
    user = crud.get_user_by_identifier(user.identifier)
    user.password = "123"
    # reset password
    status_code, data = au.reset_password(client_auth_verified, user)
    assert status_code == 500
    assert "Unable to reset your password!" in data["message"]
    assert "does not meet requirements" in data["message"]


# ###################
# UPDATE USER PROFILE
# ###################


# test updating normal attributes from user profile
def test_update_user_profile_simple_attributes(client_auth):
    # create and signin user
    user = au.create_and_signin_user(client_auth)
    # prep data
    data = {
        "email": "new_email@asreview.nl",
        "name": "new_name",
        "affiliation": "new_affiliation",
        "public": int(not user.public),
    }
    # call update
    status_code, data = au.update_user(client_auth, data)
    assert status_code == 200
    assert data["message"] == "User profile updated."


# test updating the password
def test_update_password(client_auth):
    # create and signin user
    user = au.create_and_signin_user(client_auth)
    # prep data
    new_password = "NewPassword123#"
    data = {
        "email": user.email,
        "name": user.name,
        "affiliation": user.affiliation,
        "public": int(user.public),
        "password": new_password,
    }
    # call update
    status_code, data = au.update_user(client_auth, data)
    assert status_code == 200
    assert data["message"] == "User profile updated."
    # Checking if new password works signout
    au.signout_user(client_auth)
    # signin with new password
    user.password = new_password
    status_code, data = au.signin_user(client_auth, user)
    assert status_code == 200


# test updating wrong new attribute values
@pytest.mark.parametrize(
    "attribute_data",
    [
        ("email", "email"),
        ("email", "user2@asreview.nl"),
        ("name", ""),
        ("password", "abc"),
    ],
)
def test_update_user_with_wrong_values(client_auth, attribute_data):
    # make sure I have another user to test email duplication
    user = crud.create_user(DB, 2)
    # get attribute and value from parametrize
    attr, wrong_value = attribute_data
    # create user and signin user
    user = au.create_and_signin_user(client_auth)
    data = {
        "email": user.email,
        "name": user.name,
        "affiliation": user.affiliation,
        "public": int(user.public),
        "password": "ABcd!1234",  # valid password
    }
    # manipulate attribute
    data[attr] = wrong_value
    # update
    status_code, data = au.update_user(client_auth, data)
    assert status_code == 500
    assert "Unable to update your profile" in data["message"]
    assert (attr.capitalize() in data["message"]) or attr in data["message"]


# ###################
# REFRESH
# ###################


# Test refresh: user signed in
def test_refresh_with_signed_in_user(client_auth):
    # create and signin user
    user = au.create_and_signin_user(client_auth)
    # refresh
    status_code, data = au.refresh(client_auth)
    assert status_code == 200
    assert data["id"] == user.id
    assert data["logged_in"] is True
    assert data["name"] == user.name


# Test refresh: user NOT signed in
def test_refresh_with_signed_out_user(client_auth):
    # create and signin user
    au.create_and_signin_user(client_auth)
    # signout
    au.signout_user(client_auth)
    # refresh
    status_code, data = au.refresh(client_auth)
    assert status_code == 200
    assert data["id"] is None
    assert data["logged_in"] is False
    assert data["name"] == ""


# ###################
# TEST LOGIN REQUIRED
# ###################


# User must be logged in, in order to signout,
# we expect an error if we sign out if not signed in
@pytest.mark.parametrize("api_call", [au.signout_user, au.get_profile, au.update_user])
def test_must_be_signed_in_to_signout(client_auth, api_call):
    if len(getfullargspec(api_call).args) == 1:
        status_code, data = api_call(client_auth)
    else:
        status_code, data = api_call(client_auth, {})
    # asserts
    assert status_code == 401
    assert data["message"] == "Login required."



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sklearn.ensemble import RandomForestClassifier as SKRandomForestClassifier

from asreview.models.classifiers.base import BaseTrainClassifier
from asreview.models.classifiers.utils import _set_class_weight


class RandomForestClassifier(BaseTrainClassifier):
    """
    Random forest classifier (``rf``).

    The Random Forest classifier is an implementation based
    on the sklearn Random Forest classifier.

    Arguments
    ---------
    n_estimators : int, default=100
        The number of trees in the forest.
    max_features: int, default=10
        Number of features in the model.
    class_weight: float, default=1.0
        Class weight of the inclusions.
    random_state : int or RandomState, default=None
        Controls both the randomness of the bootstrapping of the samples used
        when building trees and the sampling of the features to consider when
        looking for the best split at each node.
    """

    name = "rf"
    label = "Random forest"

    def __init__(
        self, n_estimators=100, max_features=10, class_weight=1.0, random_state=None
    ):
        super(RandomForestClassifier, self).__init__()
        self.n_estimators = int(n_estimators)
        self.max_features = int(max_features)
        self.class_weight = class_weight
        self._random_state = random_state

        self._model = SKRandomForestClassifier(
            n_estimators=self.n_estimators,
            max_features=self.max_features,
            class_weight=_set_class_weight(class_weight),
            random_state=random_state,
        )

    def full_hyper_space(self):
        from hyperopt import hp

        hyper_choices = {}
        hyper_space = {
            "mdl_n_estimators": hp.quniform("mdl_n_estimators", 10, 100, 1),
            "mdl_max_features": hp.quniform("mdl_max_features", 6, 10, 1),
            "mdl_class_weight": hp.lognormal("mdl_class_weight", 0, 1),
        }
        return hyper_space, hyper_choices



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from base64 import b64decode
from base64 import b64encode
from collections import OrderedDict
from datetime import datetime
from io import BytesIO

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse import load_npz
from scipy.sparse import save_npz

from asreview.settings import ASReviewSettings
from asreview.state.legacy.base import BaseState


# from asreview.analysis.statistics DEPRECATED
def _get_labeled_order(state):
    """Get the order in which papers were labeled."""
    label_order = []
    n_initial = 0
    n_predictor_models = state.n_predictor_models
    for query_i in range(n_predictor_models):
        try:
            label_methods = state.get("label_methods", query_i)
        except KeyError:
            continue
        label_idx = state.get("label_idx", query_i)
        for i in range(len(label_idx)):
            if label_methods[i] == "initial":
                n_initial += 1
        label_order.extend(label_idx)
    return label_order, n_initial


# from asreview.analysis.statistics DEPRECATED
def _get_last_proba_order(state):
    """Get the ranking of papers in the last query."""
    n_predictor_models = state.n_predictor_models
    pool_idx = None
    for query_i in reversed(range(n_predictor_models)):
        try:
            pool_idx = state.get("pool_idx", query_i)
        except KeyError:
            continue
        if pool_idx is not None:
            proba = state.get("proba", query_i)
            break

    if pool_idx is None:
        return []
    return pool_idx[np.argsort(-proba[pool_idx])]


def get_serial_list(array, dtype=None):
    if isinstance(array, np.ndarray):
        array = array.tolist()
    if dtype is not None:
        array = list(map(dtype, array))
    return array


class DictState(BaseState):
    """Class for storing the state of a review with no permanent storage."""

    version = "1.1"
    read_only = False

    def __init__(self, state_fp, *_, **__):
        super(DictState, self).__init__(state_fp)
        self.read_only = False

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        self.close()

    def __str__(self):
        return self._print_state()

    def _print_state(self):
        self._state_dict["time"]["end_time"] = str(datetime.now())
        label_order, _ = _get_labeled_order(self)
        try:
            labels_assigned = self.get("labels")[label_order]
        except (KeyError, IndexError):
            return ""
        labels = list(zip(label_order, labels_assigned))

        state_str = "Labeled during review:\n\n"
        for label in labels:
            state_str += f"{label[0]} => {label[1]}\n"

        pool_order = _get_last_proba_order(self)
        if len(pool_order) > 0:
            state_str += "\n\n Most likely included according to ASReview:\n\n"
            for idx in pool_order:
                state_str += f"{idx}\n"

        return state_str

    def _add_to_state(self, new_dict, i, append_result=False):
        # Find the first number that has not yet been used.
        results = self._state_dict["results"]
        if i is None:
            if len(results) > 0 and set(new_dict.keys()).isdisjoint(results[-1]):
                i = len(results) - 1
            else:
                i = len(results)

        while i >= len(results):
            results.append({})

        for key in new_dict:
            if key in results[i] and append_result:
                results[i][key].extend(new_dict[key])
            else:
                results[i][key] = new_dict[key]

    def _add_as_data(self, as_data, feature_matrix=None):
        record_table = as_data.record_ids
        data_hash = as_data.hash()

        if "data_properties" not in self._state_dict:
            self._state_dict["data_properties"] = {}

        data_properties = self._state_dict["data_properties"]
        if data_hash not in data_properties:
            data_properties[data_hash] = {}
        data_properties[data_hash]["record_table"] = record_table.tolist()
        if feature_matrix is None:
            return
        if isinstance(feature_matrix, np.ndarray):
            encoded_X = feature_matrix.tolist()
            matrix_type = "ndarray"
        elif isinstance(feature_matrix, csr_matrix):
            with BytesIO() as f:
                save_npz(f, feature_matrix)
                f.seek(0)
                encoded_X = b64encode(f.read()).decode("ascii")
            matrix_type = "csr_matrix"
        else:
            encoded_X = feature_matrix
            matrix_type = "unknown"
        data_properties[data_hash]["feature_matrix"] = encoded_X
        data_properties[data_hash]["matrix_type"] = matrix_type

    def get_feature_matrix(self, data_hash):
        my_data = self._state_dict["data_properties"][data_hash]
        encoded_X = my_data["feature_matrix"]
        matrix_type = my_data["matrix_type"]
        if matrix_type == "ndarray":
            return np.array(encoded_X)
        elif matrix_type == "csr_matrix":
            with BytesIO(b64decode(encoded_X)) as f:
                return load_npz(f)
        return encoded_X

    def get_current_queries(self):
        str_queries = self._state_dict["current_queries"]
        return {int(key): value for key, value in str_queries.items()}

    def set_current_queries(self, current_queries):
        self._state_dict["current_queries"] = current_queries

    def is_empty(self):
        return len(self._state_dict["results"]) == 0

    def set_labels(self, y):
        self._state_dict["labels"] = y.tolist()

    def set_final_labels(self, y):
        self._state_dict["final_labels"] = y.tolist()

    @property
    def settings(self):
        settings = self._state_dict.get("settings", None)
        if settings is None:
            return None
        return ASReviewSettings(**settings)

    @settings.setter
    def settings(self, settings):
        self._state_dict["settings"] = vars(settings)

    def add_classification(self, idx, labels, methods, query_i):
        # Ensure that variables are serializable
        idx = get_serial_list(idx, int)
        labels = get_serial_list(labels, int)
        methods = get_serial_list(methods, str)

        new_dict = {"labelled": list(zip(idx, labels, methods))}
        self._add_to_state(new_dict, query_i, append_result=True)

    def add_proba(self, pool_idx, train_idx, proba, query_i):
        new_dict = {
            "pool_idx": get_serial_list(pool_idx, int),
            "train_idx": get_serial_list(train_idx, int),
            "proba": get_serial_list(proba, float),
        }
        self._add_to_state(new_dict, query_i)

    def n_queries(self):
        return len(self._state_dict["results"])

    def get(self, variable, query_i=None, idx=None):
        if query_i is not None:
            res = self._state_dict["results"][query_i]

        array = None
        label_vars = ["label_idx", "inclusions", "label_methods"]
        if variable in label_vars:
            array_id = label_vars.index(variable)
            if variable == "label_methods":
                dtype = str
            else:
                dtype = int
            array = np.array([x[array_id] for x in res["labelled"]], dtype=dtype)
        elif variable == "labels":
            array = np.array(self._state_dict["labels"], dtype=int)
        elif variable == "final_labels":
            array = np.array(self._state_dict["final_labels"], dtype=int)
        elif variable == "proba":
            array = np.array(res["proba"], dtype=float)
        elif variable == "train_idx":
            array = np.array(res["train_idx"], dtype=int)
        elif variable == "pool_idx":
            array = np.array(res["pool_idx"], dtype=int)
        if array is None:
            return None
        if idx is not None:
            return array[idx]
        return array

    def delete_last_query(self):
        self._state_dict["results"].pop()

    def initialize_structure(self):
        from asreview import __version__ as asr_version

        self._state_dict = OrderedDict(
            {
                "time": {"start_time": str(datetime.now())},
                "version": self.version,
                "software_version": asr_version,
                "settings": {},
                "results": [],
            }
        )

    def close(self):
        if not self.read_only:
            self.save()

    def save(self):
        print(self)

    def restore(self, *_, **__):
        self.initialize_structure()



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Random sampling strategy."""

import numpy as np

from asreview.models.query.base import BaseQueryStrategy
from asreview.utils import get_random_state


class RandomQuery(BaseQueryStrategy):
    """Random query strategy (``random``).

    Randomly select samples with no regard to model assigned probabilities.

    .. warning::
        Selecting this option means your review is not going to be
        accelerated by ASReview.

    """

    name = "random"
    label = "Random"

    def __init__(self, random_state=None):
        super(RandomQuery, self).__init__()
        self._random_state = get_random_state(random_state)

    def query(self, X, classifier=None, n_instances=None, **kwargs):
        if n_instances is None:
            n_instances = X.shape[0]

        query_idx = self._random_state.choice(
            np.arange(X.shape[0]), n_instances, replace=False
        )

        return query_idx



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from sentence_transformers import models
    from sentence_transformers.SentenceTransformer import SentenceTransformer
except ImportError:
    ST_AVAILABLE = False
else:
    ST_AVAILABLE = True

from asreview.models.feature_extraction.base import BaseFeatureExtraction


def _check_st():
    if not ST_AVAILABLE:
        raise ImportError("Install sentence-transformers package to use Sentence BERT.")


class SBERT(BaseFeatureExtraction):
    """Sentence BERT feature extraction technique (``sbert``).

    By setting the ``transformer_model`` parameter, you can use other
    transformer models. For example, ``transformer_model='bert-base-nli-stsb-
    large'``. For a list of available models, see the `Sentence BERT
    documentation <https://huggingface.co/sentence-transformers>`__.

    Sentence BERT is a sentence embedding model that is trained on a large
    corpus of human written text. It is a fast and accurate model that can
    be used for many tasks.

    The huggingface library includes multilingual text classification models. If
    your dataset contains records with multiple languages, you can use the
    ``transformer_model`` parameter to select the model that is most suitable
    for your data.

    .. note::

        This feature extraction technique requires ``sentence_transformers``
        to be installed. Use ``pip install sentence_transformers`` or install
        all optional ASReview dependencies with ``pip install asreview[all]``
        to install the package.

    Parameters
    ----------
    transformer_model : str, optional
        The transformer model to use.
        Default: 'all-mpnet-base-v2'
    is_pretrained_SBERT: boolean, optional
        Default: True
    pooling_mode: str, optional
        Pooling mode to get sentence embeddings from word embeddings
        Default: 'mean'
        Other options available are 'mean', 'max' and 'cls'.
        Only used if is_pretrained_SBERT=False
        mean: Uses mean pooling of word embeddings
        max: Uses max pooling of word embeddings
        cls: Uses embeddings of [CLS] token as sentence embeddings
    """

    name = "sbert"
    label = "Sentence BERT"

    def __init__(
        self,
        *args,
        transformer_model="all-mpnet-base-v2",
        is_pretrained_sbert=True,
        pooling_mode="mean",
        **kwargs
    ):
        super(SBERT, self).__init__(*args, **kwargs)
        self.transformer_model = transformer_model
        self.is_pretrained_sbert = is_pretrained_sbert
        self.pooling_mode = pooling_mode

    def transform(self, texts):
        _check_st()

        if self.is_pretrained_sbert:
            model = SentenceTransformer(self.transformer_model)
        else:
            # If transformer_model is not a pretrained sentence transformer model,
            # add a pooling layer to get the pooled sentence embeddings from the
            # word embeddings
            word_embedding_model = models.Transformer(self.transformer_model)
            pooling_layer = models.Pooling(
                word_embedding_model.get_word_embedding_dimension(),
                pooling_mode=self.pooling_mode,
            )
            model = SentenceTransformer(modules=[word_embedding_model, pooling_layer])
        print("Encoding texts using sbert, this may take a while...")
        X = model.encode(texts, show_progress_bar=True)

        return X



import json
from argparse import Namespace
from pathlib import Path
from unittest.mock import patch
from uuid import uuid4

import pytest

import asreview.entry_points.auth_tool as tool
from asreview.entry_points.auth_tool import AuthTool
from asreview.state.sql_converter import upgrade_asreview_project_file
from asreview.utils import asreview_path
from asreview.webapp import DB
from asreview.webapp.tests.utils import api_utils as au
from asreview.webapp.tests.utils import config_parser as cp
from asreview.webapp.tests.utils import crud
from asreview.webapp.tests.utils import misc


def get_auth_tool_object(namespace):
    """This function returns an AuthTool object in which
    the session feature is set to the test database and
    the args feature is set with the <namespace> input
    parameter."""
    tool = AuthTool()
    # manipulate the DB session
    tool.session = DB.session
    tool.args = namespace
    return tool


def interactive_user_data():
    """This function returns a list of strings that can be
    used to trigger an interactively created user account."""
    user_data = cp.get_user_data(1)
    return [
        "Y",
        user_data["email"],
        user_data["name"],
        user_data["affiliation"],
        user_data["password"],
        "n",
    ]


def import_2_unauthenticated_projects(with_upgrade=True):
    """This function retrieves 2 zipped project (version 0.x)
    files from github and copies them in the asreview folder.
    To use them in tests they need to be upgraded. Both projects
    are returned."""
    # get 2 unauthenticated projects
    url1 = misc.retrieve_project_url_github("v0.19")
    url2 = misc.retrieve_project_url_github("v0.18")
    # import projects
    proj1 = misc.copy_github_project_into_asreview_folder(url1)
    proj2 = misc.copy_github_project_into_asreview_folder(url2)
    if with_upgrade:
        # update these projects to a 1.x-ish config
        upgrade_asreview_project_file(proj1.project_path)
        upgrade_asreview_project_file(proj2.project_path)
    return proj1, proj2


# Test verifying uuid: correct
def test_verify_id_correct_id():
    id = uuid4().hex
    assert tool.verify_id(id)


# Test verifying uuid: incorrect id
def test_verify_id_incorrect_id():
    id = "incorrect-id"
    assert not tool.verify_id(id)


# Test inserting a user into the database
def test_insert_user(client_auth):
    # count users
    assert crud.count_users() == 0
    # get some user credentials
    user_data = cp.get_user_data(1)
    # insert the returned dictionary
    tool.insert_user(DB.session, user_data)
    # count users again
    assert crud.count_users() == 1
    # get user
    user = crud.last_user()
    assert user.email == user_data["email"]
    assert user.identifier == user_data["email"]
    assert user.origin == "asreview"
    assert user.name == user_data["name"]
    assert user.affiliation == user_data["affiliation"]
    assert user.confirmed


# Test inserting a duplicate
def test_insert_user_duplicate(client_auth):
    # count users
    assert crud.count_users() == 0
    # get some user credentials
    user_data = cp.get_user_data(1)
    # insert the returned dictionary
    tool.insert_user(DB.session, user_data)
    # verify user has been created
    assert crud.count_users() == 1
    # and again
    result = tool.insert_user(DB.session, user_data)
    # asserts
    assert not result
    # no inserts, count remains 1
    assert crud.count_users() == 1


# Test inserting a project record in the database
def test_inserting_a_project_record(client_auth):
    # count projects
    assert crud.count_projects() == 0
    # insert this data
    data = {"project_id": uuid4().hex, "owner_id": 2}
    tool.insert_project(DB.session, data)
    # count again
    assert crud.count_projects() == 1
    # get last record
    project = crud.last_project()
    assert project.project_id == data["project_id"]
    assert project.owner_id == data["owner_id"]


# Test updating a project record in the database
def test_updating_a_project_record(client_auth):
    # count projects
    assert crud.count_projects() == 0
    # insert this data
    data = {"project_id": uuid4().hex, "owner_id": 2}
    tool.insert_project(DB.session, data)
    # count again
    assert crud.count_projects() == 1
    # change owner id
    data["owner_id"] = 3
    tool.insert_project(DB.session, data)
    # count again, no inserts, count remains 1
    assert crud.count_projects() == 1
    project = crud.last_project()
    assert project.project_id == data["project_id"]
    assert project.owner_id == data["owner_id"]


# Test get users
def test_get_users(client_auth):
    # create 2 users
    user1 = crud.create_user(DB, 1)
    user2 = crud.create_user(DB, 2)
    print(user1, user2)
    assert crud.count_users() == 2
    # test function
    result = tool.get_users(DB.session)
    assert set(result) == set([user1, user2])


# ####################
# Test AuthTool Object
# ####################


# insert users with json string
def test_auth_tool_add_users_with_json(client_auth):
    # assert we have no users
    assert crud.count_users() == 0
    # build appropriate json argument
    user_data = cp.get_user_data(1)
    namespace = Namespace(json=f"{json.dumps([user_data])}")
    # get auth_tool object
    auth_tool = get_auth_tool_object(namespace)
    # execute add users function
    auth_tool.add_users()
    # assert we now have a user
    assert crud.count_users() == 1
    user = crud.last_user()
    # quick asserts, we have already tested this in
    # test_insert_user
    assert user.email == user_data["email"]
    assert user.identifier == user_data["email"]


# insert users interactively, correct input
def test_auth_tool_add_users_interact(client_auth):
    # assert we have no users
    assert crud.count_users() == 0
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # build interactive input
    answers = interactive_user_data()
    with patch('builtins.input', side_effect=answers):
        auth_tool.add_users()
    # assert we now have a users
    assert crud.count_users() == 1


# insert users interactively, incorrect email
def test_auth_tool_add_users_interact_incorr_email(client_auth, capsys):
    """This test interactively inserts a user account, but
    a non valid email address is provided"""
    # assert we have no users
    assert crud.count_users() == 0
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # build interactive input
    answers = interactive_user_data()
    # add in a faulty email address
    answers.insert(1, "abcd@")
    with patch('builtins.input', side_effect=answers):
        auth_tool.add_users()
    _, err = capsys.readouterr()
    assert "Entered email address is not recognized" in err
    # assert we now have a users
    assert crud.count_users() == 1


# insert users interactively, name too short
def test_auth_tool_add_users_interact_incorr_name(client_auth, capsys):
    """This test interactively inserts a user account, but
    a short name is provided"""
    # assert we have no users
    assert crud.count_users() == 0
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # build interactive input
    answers = interactive_user_data()
    # add in a name that is too short
    answers.insert(2, "ab")
    with patch('builtins.input', side_effect=answers):
        auth_tool.add_users()
    _, err = capsys.readouterr()
    assert "Full name must contain more than 2" in err
    # assert we now have a users
    assert crud.count_users() == 1


# insert users interactively, bad password
def test_auth_tool_add_users_interact_incorr_passw(client_auth, capsys):
    """This test interactively inserts a user account, but
    a non valid password is provided"""
    # assert we have no users
    assert crud.count_users() == 0
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # build interactive input
    answers = interactive_user_data()
    # add in a name that is too short
    answers.insert(4, "1111")
    with patch('builtins.input', side_effect=answers):
        auth_tool.add_users()
    _, err = capsys.readouterr()
    assert "Use 8 or more characters with a mix" in err
    # assert we now have a users
    assert crud.count_users() == 1


# Test validity check. Note: this and the next test can not
# be parametrized because of the tested function: it -needs-
# to be finished with a correct value
def test_validity_function_valid(capsys):
    """Tests the _ensure_valid_value_for method, expects
    no error messages if the input value respects the
    lambda function."""
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # define a correct value
    correct = "a"
    hint = "Test hint"
    # run function with patched input
    with patch('builtins.input', side_effect=[correct]):
        # run validity function
        auth_tool._ensure_valid_value_for(
            "test", lambda x: x == correct, hint=hint
        )
    out, err = capsys.readouterr()
    assert not bool(out)
    assert not bool(err)


# Test validity check, see remark previous test if you notice
# the repetition
def test_validity_function_invalid(capsys):
    """Tests the _ensure_valid_value_for method, expects
    error messages if the input value does not respect the
    lambda function."""
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # define a correct value
    correct = "a"
    incorrect = "b"
    hint = "Test hint"
    # run function with patched input
    with patch('builtins.input', side_effect=[incorrect, correct]):
        # run validity function
        auth_tool._ensure_valid_value_for(
            "test", lambda x: x == correct, hint=hint
        )
    out, err = capsys.readouterr()
    assert not bool(out)
    assert err == hint


# Test printing a project
def test_print_project(capsys):
    keys = ["folder", "version", "project_id", "name", "authors", "created"]
    data = {k: uuid4().hex for k in keys}
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # run function
    auth_tool._print_project(data)
    out, _ = capsys.readouterr()
    assert f"* {data['folder']}" in out
    assert f"version: {data['version']}" in out
    assert f"id: {data['project_id']}" in out
    assert f"name: {data['name']}" in out
    assert f"authors: {data['authors']}" in out
    assert f"created: {data['created']}" in out


# Test printing a user with affiliation
def test_print_user_with_affiliation(client_auth, capsys):
    user = crud.create_user(DB, 1)
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # run function
    auth_tool._print_user(user)
    out, _ = capsys.readouterr()
    expected = f"{user.id} - {user.email} ({user.name}), {user.affiliation}"
    assert out.strip() == expected


# Test printing a user without affiliation
def test_print_user_without_affiliation(client_auth, capsys):
    user = crud.create_user(DB, 1)
    user.affiliation = None
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # run function
    auth_tool._print_user(user)
    out, _ = capsys.readouterr()
    expected = f"{user.id} - {user.email} ({user.name})"
    assert out.strip() == expected


# Testing _get_projects
def test_get_projects(client_no_auth):
    # create a project
    _, data = au.create_project(client_no_auth, "test")
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # run function
    result = auth_tool._get_projects()
    assert isinstance(result, list)
    assert len(result) == 1
    result = result[0]
    assert result["folder"] == data["id"]
    assert result["version"] == data["version"]
    assert result["project_id"] == data["id"]
    assert result["name"] == data["name"]
    assert result["authors"] == data["authors"]
    assert result["created"] == data["datetimeCreated"]
    assert result["owner_id"] == 0


# Test listing users
def test_list_users(client_auth, capsys):
    # create 2 users
    u1 = crud.create_user(DB, 1)
    u2 = crud.create_user(DB, 2)
    assert crud.count_users() == 2
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # run function
    auth_tool.list_users()
    out, _ = capsys.readouterr()
    exp1 = f"{u1.id} - {u1.email} ({u1.name}), {u1.affiliation}"
    exp2 = f"{u2.id} - {u2.email} ({u2.name}), {u2.affiliation}"
    assert exp1 in out
    assert exp2 in out


# Test list projects: no json data
def test_list_projects_no_json(client_no_auth, capsys):
    # create two projects
    _, data1 = au.create_project(client_no_auth, "test1")
    _, data2 = au.create_project(client_no_auth, "test2")
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # run function
    auth_tool.list_projects()
    out, _ = capsys.readouterr()
    # we have already tested _print_project, so I will keep
    # it short
    assert f"* {data1['id']}" in out
    assert f"* {data2['id']}" in out
    assert f"name: {data1['name']}" in out
    assert f"name: {data2['name']}" in out


# Test list projects: output is a json string
def test_list_projects_with_json(client_no_auth, capsys):
    # create two projects
    _, data1 = au.create_project(client_no_auth, "test1")
    _, data2 = au.create_project(client_no_auth, "test2")
    data = {
        data1.get("id"): data1,
        data2.get("id"): data2
    }
    # get auth_tool object
    auth_tool = get_auth_tool_object(Namespace(json=True))
    # run function
    auth_tool.list_projects()
    out, _ = capsys.readouterr()
    # this loads the out json string into a list of dicts
    out = json.loads(json.loads(out))
    assert isinstance(out, list)
    assert len(out) == 2
    for proj in out:
        expected = data[proj["project_id"]]
        assert proj["folder"] == expected["id"]
        assert proj["version"] == expected["version"]
        assert proj["project_id"] == expected["id"]
        assert proj["name"] == expected["name"]
        assert proj["authors"] == expected["authors"]
        assert proj["created"] == expected["datetimeCreated"]
        assert proj["owner_id"] == 0


# Test linking projects to users with a json string
# Note: We can not simulate a conversion from an unauthenticated
# app into an authenticated one. To overcome this problem, 2 old
# project zip files (version 0.x) are copied from Github into the
# asreview folder and upgraded. This is done without the help of
# the API, ensuring they can't be linked to a User account.
def test_link_project_with_json_string(client_auth, capsys):
    # import projects
    proj1, proj2 = import_2_unauthenticated_projects()
    # create 2 users
    user1 = crud.create_user(DB, 1)
    user2 = crud.create_user(DB, 2)
    # check database
    assert crud.count_users() == 2
    assert crud.count_projects() == 0
    # check if we have 2 folders in asreview path
    assert len(misc.get_folders_in_asreview_path()) == 2
    # get from the auth tool a json string
    auth_tool = get_auth_tool_object(Namespace(json=True))
    auth_tool.list_projects()
    out, _ = capsys.readouterr()
    # we replace the owner ids with the ids of the users
    json_string = out.replace(": 0", f": {user1.id}", 1)
    json_string = json_string.replace(": 0", f": {user2.id}", 1)

    # use this string to run the function with a new AuthTool
    auth_tool = get_auth_tool_object(Namespace(json=json.loads(json_string)))
    auth_tool.link_projects()
    # check database and check if the users own the correct project
    assert crud.count_projects() == 2
    project_dict = {
        proj["owner_id"]: proj
        for proj in json.loads(json.loads(json_string))
    }
    for user in [user1, user2]:
        expected_proj = project_dict[user.id]
        assert len(user.projects) == 1
        assert user.projects[0].project_id == expected_proj["folder"]
        assert user.projects[0].project_id == expected_proj["project_id"]
        # check also on the file-system
        assert Path(asreview_path() / expected_proj["folder"]).exists()


# Test linking projects interactively
def test_link_projects_interactively(client_auth):
    # import projects
    proj1, proj2 = import_2_unauthenticated_projects()
    project_data = {p.config.get("id"): p for p in [proj1, proj2]}
    # create a user
    user = crud.create_user(DB, 1)
    # check the database
    assert crud.count_users() == 1
    assert crud.count_projects() == 0
    # create AuthTool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # run function with patched input
    with patch('builtins.input', side_effect=[user.id, user.id]):
        # link project to user
        auth_tool.link_projects()
    # check database again
    assert crud.count_projects() == 2
    # make sure the user has 2 different projects
    assert len([p.project_id for p in user.projects]) == 2
    # check user projects
    for project in user.projects:
        org_data = project_data[project.project_id]
        assert org_data.config.get("id") == project.project_id
        assert Path(asreview_path() / project.project_id).exists()


# Test linking projects with a typo
def test_link_projects_interactively_with_typo(client_auth):
    # import projects
    proj1, proj2 = import_2_unauthenticated_projects()
    # create a user
    user = crud.create_user(DB, 1)
    # check the database
    assert crud.count_users() == 1
    assert crud.count_projects() == 0
    # create AuthTool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # run function with patched input (there is a wrong id in there)
    with patch('builtins.input', side_effect=[user.id, str(-5), user.id]):
        # link project to user
        auth_tool.link_projects()
    # check database again
    assert crud.count_projects() == 2


# Test failure of anything related to projects if a project is older
# than version 0.x.
@pytest.mark.parametrize(
    "method",
    [
        "_generate_project_links",
        "list_projects",
        "link_projects"
    ]
)
def test_projects_with_0x_projects(client_auth, method):
    # import projects
    proj1, proj2 = import_2_unauthenticated_projects(with_upgrade=False)
    # make sure these projects exist
    assert len(misc.get_folders_in_asreview_path()) == 2
    # create AuthTool object
    auth_tool = get_auth_tool_object(Namespace(json=None))
    # try to link project to user
    with pytest.raises(RuntimeError) as error:
        func = getattr(auth_tool, method)
        func()
        assert "Version of project with id" in str(error.value)
        assert "too old" in str(error.value)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from sklearn.cluster import KMeans

from asreview.models.query.base import ProbaQueryStrategy
from asreview.models.query.max import MaxQuery
from asreview.utils import get_random_state


class ClusterQuery(ProbaQueryStrategy):
    """Clustering query strategy (``cluster``).

    Use clustering after feature extraction on the dataset. Then the highest
    probabilities within random clusters are sampled.

    Arguments
    ---------
    cluster_size: int
        Size of the clusters to be made. If the size of the clusters is
        smaller than the size of the pool, fall back to max sampling.
    update_interval: int
        Update the clustering every x instances.
    random_state: int, RandomState
        State/seed of the RNG.
    """

    name = "cluster"
    label = "Clustering"

    def __init__(self, cluster_size=350, update_interval=200, random_state=None):
        """Initialize the clustering strategy."""
        super(ClusterQuery, self).__init__()
        self.cluster_size = cluster_size
        self.update_interval = update_interval
        self.last_update = None
        self.fallback_model = MaxQuery()
        self._random_state = get_random_state(random_state)

    def _query(self, predictions, n_instances, X):
        n_samples = X.shape[0]

        last_update = self.last_update
        if (
            last_update is None
            or self.update_interval is None
            or last_update - n_samples >= self.update_interval
        ):
            n_clusters = round(n_samples / self.cluster_size)
            if n_clusters <= 1:
                return self.fallback_model._query(predictions, n_instances, X)
            model = KMeans(
                n_clusters=n_clusters, n_init=1, random_state=self._random_state
            )
            self.clusters = model.fit_predict(X)
            self.last_update = n_samples

        clusters = {}
        for idx in np.arange(n_samples):
            cluster_id = self.clusters[idx]
            if cluster_id in clusters:
                clusters[cluster_id].append((idx, predictions[idx, 1]))
            else:
                clusters[cluster_id] = [(idx, predictions[idx, 1])]

        for cluster_id in clusters:
            try:
                clusters[cluster_id] = sorted(clusters[cluster_id], key=lambda x: x[1])
            except ValueError:
                raise

        clust_idx = []
        cluster_ids = list(clusters)
        for _ in range(n_instances):
            cluster_id = self._random_state.choice(cluster_ids, 1)[0]
            clust_idx.append(clusters[cluster_id].pop()[0])
            if len(clusters[cluster_id]) == 0:
                del clusters[cluster_id]
                cluster_ids = list(clusters)

        clust_idx = np.array(clust_idx, dtype=int)

        return clust_idx

    def full_hyper_space(self):
        from hyperopt import hp

        parameter_space = {
            "qry_cluster_size": hp.quniform("qry_cluster_size", 50, 1000, 1),
            "qry_update_interval": hp.quniform("qry_update_interval", 100, 300, 1),
        }
        return parameter_space, {}



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.models.classifiers.logistic import LogisticClassifier
from asreview.models.classifiers.lstm_base import LSTMBaseClassifier
from asreview.models.classifiers.lstm_pool import LSTMPoolClassifier
from asreview.models.classifiers.nb import NaiveBayesClassifier
from asreview.models.classifiers.nn_2_layer import NN2LayerClassifier
from asreview.models.classifiers.rf import RandomForestClassifier
from asreview.models.classifiers.svm import SVMClassifier
from asreview.models.classifiers.utils import get_classifier
from asreview.models.classifiers.utils import get_classifier_class
from asreview.models.classifiers.utils import list_classifiers

"""Machine learning classifiers to classify the documents.

There are several machine learning classifiers available. In configuration
files, parameters are found under the section ``[model_param]``.
"""



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

try:
    import tensorflow as tf
    from tensorflow.keras import regularizers
    from tensorflow.keras.layers import Dense
    from tensorflow.keras.models import Sequential
except ImportError:
    TF_AVAILABLE = False
else:
    TF_AVAILABLE = True
    try:
        tf.logging.set_verbosity(tf.logging.ERROR)
    except AttributeError:
        logging.getLogger("tensorflow").setLevel(logging.ERROR)

import numpy as np
import scipy

from asreview.models.classifiers.base import BaseTrainClassifier
from asreview.models.classifiers.lstm_base import _get_optimizer
from asreview.models.classifiers.utils import _set_class_weight


def _check_tensorflow():
    if not TF_AVAILABLE:
        raise ImportError(
            "Install tensorflow package to use"
            " Fully connected neural network (2 hidden layers).")


class NN2LayerClassifier(BaseTrainClassifier):
    """Fully connected neural network (2 hidden layers) classifier (``nn-2-layer``).

    Neural network with two hidden, dense layers of the same size.

    Recommended feature extraction model is
    :class:`asreview.models.feature_extraction.Doc2Vec`.

    .. note::

        This model requires ``tensorflow`` to be installed. Use ``pip install
        tensorflow`` or install all optional ASReview dependencies with ``pip
        install asreview[all]``

    .. warning::

        Might crash on some systems with limited memory in
        combination with :class:`asreview.models.feature_extraction.Tfidf`.

    Arguments
    ---------
    dense_width: int
        Size of the dense layers.
    optimizer: str
        Name of the Keras optimizer.
    learn_rate: float
        Learning rate multiplier of the default learning rate.
    regularization: float
        Strength of the regularization on the weights and biases.
    verbose: int
        Verbosity of the model mirroring the values for Keras.
    epochs: int
        Number of epochs to train the neural network.
    batch_size: int
        Batch size used for the neural network.
    shuffle: bool
        Whether to shuffle the training data prior to training.
    class_weight: float
        Class weights for inclusions (1's).
    """

    name = "nn-2-layer"
    label = "Fully connected neural network (2 hidden layers)"

    def __init__(self,
                 dense_width=128,
                 optimizer='rmsprop',
                 learn_rate=1.0,
                 regularization=0.01,
                 verbose=0,
                 epochs=35,
                 batch_size=32,
                 shuffle=False,
                 class_weight=30.0):
        """Initialize the 2-layer neural network model."""
        super(NN2LayerClassifier, self).__init__()
        self.dense_width = int(dense_width)
        self.optimizer = optimizer
        self.learn_rate = learn_rate
        self.regularization = regularization
        self.verbose = verbose
        self.epochs = int(epochs)
        self.batch_size = int(batch_size)
        self.shuffle = shuffle
        self.class_weight = class_weight

        self._model = None
        self.input_dim = None

    def fit(self, X, y):
        # check is tensorflow is available
        _check_tensorflow()

        if scipy.sparse.issparse(X):
            X = X.toarray()
        if self._model is None or X.shape[1] != self.input_dim:
            self.input_dim = X.shape[1]
            self._model = _create_dense_nn_model(
                self.input_dim,
                self.dense_width,
                self.optimizer,
                self.learn_rate,
                self.regularization,
                self.verbose,
            )

        self._model.fit(
            X,
            y,
            batch_size=self.batch_size,
            epochs=self.epochs,
            shuffle=self.shuffle,
            verbose=self.verbose,
            class_weight=_set_class_weight(self.class_weight),
        )

    def predict_proba(self, X):
        if scipy.sparse.issparse(X):
            X = X.toarray()
        pos_pred = self._model.predict(X, verbose=self.verbose)
        neg_pred = 1 - pos_pred
        return np.hstack([neg_pred, pos_pred])

    def full_hyper_space(self):
        from hyperopt import hp
        hyper_choices = {
            "mdl_optimizer": ["sgd", "rmsprop", "adagrad", "adam", "nadam"]
        }
        hyper_space = {
            "mdl_dense_width": hp.quniform("mdl_dense_width", 2, 100, 1),
            "mdl_epochs": hp.quniform("mdl_epochs", 20, 60, 1),
            "mdl_optimizer": hp.choice("mdl_optimizer", hyper_choices["mdl_optimizer"]),
            "mdl_learn_rate": hp.lognormal("mdl_learn_rate", 0, 1),
            "mdl_class_weight": hp.lognormal("mdl_class_weight", 3, 1),
            "mdl_regularization": hp.lognormal("mdl_regularization", -4, 2),
        }
        return hyper_space, hyper_choices


def _create_dense_nn_model(vector_size=40,
                           dense_width=128,
                           optimizer='rmsprop',
                           learn_rate_mult=1.0,
                           regularization=0.01,
                           verbose=1):
    """Return callable model.

    Returns
    -------
    callable:
        A function that return the Keras Sklearn model when
        called.

    """

    # check is tensorflow is available
    _check_tensorflow()

    model = Sequential()

    model.add(
        Dense(
            dense_width,
            input_dim=vector_size,
            kernel_regularizer=regularizers.l2(regularization),
            activity_regularizer=regularizers.l1(regularization),
            activation='relu',
        ))

    # add Dense layer with relu activation
    model.add(
        Dense(
            dense_width,
            kernel_regularizer=regularizers.l2(regularization),
            activity_regularizer=regularizers.l1(regularization),
            activation='relu',
        ))

    # add Dense layer
    model.add(Dense(1, activation='sigmoid'))

    optimizer_fn = _get_optimizer(optimizer, learn_rate_mult)

    # Compile model
    model.compile(
        loss='binary_crossentropy',
        optimizer=optimizer_fn,
        metrics=['acc'])

    if verbose >= 1:
        model.summary()

    return model



import io
import json
import random
import re
from pathlib import Path
from typing import Union
from urllib.request import urlopen

import requests
from flask import current_app

from asreview.project import ASReviewProject
from asreview.utils import asreview_path


def current_app_is_authenticated():
    return current_app.config.get("AUTHENTICATION_ENABLED")


def get_project_id(project):
    """Get a project id from either a Project model
    (authenticated app) or an ASReviewProject object
    (unauthenticated app)."""
    id = None
    if current_app_is_authenticated():
        id = project.project_id
    else:
        id = project.config["id"]
    return id


def read_project_file(project):
    """Loads the data from the project.json file."""
    id = get_project_id(project)
    with open(asreview_path() / id / "project.json", "r") as f:
        data = json.load(f)
        return data


def manipulate_project_file(project, key, value):
    """Updates key value pairs in the project.json file."""
    id = get_project_id(project)
    data = read_project_file(project)
    data[key] = value
    with open(asreview_path() / id / "project.json", "w+") as f:
        json.dump(data, f)
        return True
    return False


def _extract_stem(path: Union[str, Path]):
    """Extracts a stem from a path or URL containing a filename."""
    return Path(re.split(":|/", str(path))[-1]).stem


def extract_filename_stem(upload_data):
    """Helper function to get the stem part of a filename from a
    Path or URL contaning a filename."""
    # upload data is a dict with a single key value pair
    value = list(upload_data.values())[0]
    # split this value on either / or :
    return _extract_stem(value)


def choose_project_algorithms():
    """Randomly chooses a model plus the appropriate feature
    extraction, query strategy and balance strategy."""
    model = random.choice(["svm", "nb", "logistic"])
    feature_extraction = random.choice(["tfidf"])
    data = {
        "model": model,
        "feature_extraction": feature_extraction,
        "query_strategy": random.choice(
            ["cluster", "max", "max_random", "max_uncertainty", "random", "uncertainty"]
        ),
        "balance_strategy": random.choice(["double", "simple", "undersample"]),
    }
    return data


def retrieve_project_url_github(version=None):
    """Retrieve .asreview file(s) url from asreview-project-files-testing
    GitHub repository. When version is not None, the function resturns
    a single URL, otherwise a list containing URLs."""

    repo = "asreview/asreview-project-files-testing"
    repo_api_url = f"https://api.github.com/repos/{repo}/git/trees/master"
    repo_url = f"https://github.com/{repo}/blob/master"
    file_type = "startreview.asreview?raw=true"

    json_file = json.loads(urlopen(repo_api_url).read().decode("utf-8"))["tree"]

    version_tags = []
    project_urls = []

    for file in json_file:
        if file["type"] == "tree":
            version_tags.append(file["path"])

    for tag in version_tags:
        file_version = f"/{tag}/asreview-project-{tag.replace('.', '-')}-"
        url = repo_url + file_version + file_type

        if version is None:
            project_urls.append(url)
        else:
            return url

    return project_urls


def copy_github_project_into_asreview_folder(url):
    """This function copies a, on Github stored, ASReview project
    into the asreview folder."""
    response = requests.get(url)
    return ASReviewProject.load(
        io.BytesIO(response.content),
        asreview_path(),
        safe_import=True
    )


def get_folders_in_asreview_path():
    """This function returns the amount of folders located
    in the asreview folder."""
    return [f for f in asreview_path().glob("*") if f.is_dir()]



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod

from asreview.models.base import BaseModel


class BaseQueryStrategy(BaseModel):
    """Abstract class for query strategies."""

    name = "base-query"

    @abstractmethod
    def query(self, X, classifier=None, n_instances=None, **kwargs):
        """Query new instances.

        Arguments
        ---------
        X: numpy.ndarray
            Feature matrix to choose samples from.
        classifier: SKLearnModel
            Trained classifier to compute probabilities if they are necessary.
        n_instances: int
            Number of instances to query.

        Returns
        -------
        (numpy.ndarray, numpy.ndarray)
            The first is an array of shape (n_instances,) containing the row
            indices of the new instances in query order. The second is an array
            of shape (n_instances, n_feature_matrix_columns), containing the
            feature vectors of the new instances.
        """
        raise NotImplementedError


class ProbaQueryStrategy(BaseQueryStrategy):
    name = "proba"

    def query(self, X, classifier, n_instances=None, **kwargs):
        """Query method for strategies which use class probabilities."""
        if n_instances is None:
            n_instances = X.shape[0]

        predictions = classifier.predict_proba(X)

        query_idx = self._query(predictions, n_instances, X)

        return query_idx

    @abstractmethod
    def _query(self, predictions, n_instances, X=None):
        raise NotImplementedError



from datetime import datetime as dt
from datetime import timedelta

import pytest
from sqlalchemy.exc import IntegrityError

import asreview.webapp.tests.utils.config_parser as cp
import asreview.webapp.tests.utils.crud as crud
from asreview.webapp import DB
from asreview.webapp.authentication.models import User

# #############
# CREATE
# #############


# test identifier validation
def test_user_must_have_identifier(setup_teardown):
    user = crud.create_user(DB)
    with pytest.raises(ValueError):
        user.identifier = None

    with pytest.raises(ValueError):
        user.identifier = ""


# test uniqueness of identifier
def test_uniqueness_of_identifier(setup_teardown):
    user1 = crud.create_user(DB)
    assert crud.count_users() == 1
    # create second user with identical identifier
    user2 = cp.get_user(2)
    # set to an existing identifier
    user2.identifier = user1.identifier
    with pytest.raises(IntegrityError):
        crud.create_user(DB, user2)


# test origin validation
def test_user_must_have_origin(setup_teardown):
    user = crud.create_user(DB)
    with pytest.raises(ValueError):
        user.origin = None

    with pytest.raises(ValueError):
        user.origin = ""


# test name validation
def test_user_must_have_name(setup_teardown):
    user = crud.create_user(DB)
    with pytest.raises(ValueError, match="Name is required"):
        user.name = None

    with pytest.raises(ValueError, match="Name is required"):
        user.name = ""

    with pytest.raises(ValueError, match="Name must contain more than 2 characters"):
        user.name = "a"

    with pytest.raises(ValueError, match="Name must contain more than 2 characters"):
        user.name = "ab"


# test if email is not blank if origin is "asreview"
def test_email_validation_1(setup_teardown):
    user = crud.create_user(DB)
    user.origin = "asreview"
    with pytest.raises(ValueError, match="Email is required when origin is 'asreview'"):
        user.email = None

    with pytest.raises(ValueError, match="Email is required when origin is 'asreview'"):
        user.email = ""


# test if all fails when email is invalid
def test_email_validation_2(setup_teardown):
    user_data = crud.create_user(DB)
    invalid_email = "invalid"

    with pytest.raises(
        ValueError, match=f"Email address '{invalid_email}' is not valid"
    ):
        User(
            invalid_email,
            email=invalid_email,
            name=user_data.name,
            origin="asreview",
            password="ABCd1234!",
        )


# test uniqueness of email
def test_uniqueness_of_email(setup_teardown):
    user1 = crud.create_user(DB)
    assert crud.count_users() == 1
    # create second user with identical email
    user2 = cp.get_user(2)
    # set to an existing identifier
    user2.email = user1.email
    with pytest.raises(IntegrityError):
        crud.create_user(DB, user2)


# test if all fails when password doesn't meet requirements
@pytest.mark.parametrize(
    "password", ["", None, "a1!", "aaaaaaaaaaaaa", "1111111111111"]
)
def test_password_validation(setup_teardown, password):
    with pytest.raises(
        ValueError, match=f'Password "{str(password)}" does not meet requirements'
    ):
        User(
            "admin@asreview.nl",
            email="admin@asreview.nl",
            name="Casper",
            origin="asreview",
            password=password,
        )


# Verify we can add a user record
def test_add_user_record(setup_teardown):
    user = crud.create_user(DB)
    # verify we have 1 record
    assert crud.count_users() == 1
    assert crud.last_user() == user


# #############
# UPDATE
# #############


# Verify we can update a user record
def test_update_user_record(setup_teardown):
    user = crud.create_user(DB)
    old_hashed_password = user.hashed_password

    new_email = "new_email@asreview.nl"
    new_name = "New Name"
    new_affiliation = "New Affiliation"
    new_password = "NewPassword@123"
    new_public = False

    user.update_profile(
        email=new_email,
        name=new_name,
        affiliation=new_affiliation,
        password=new_password,
        public=new_public,
    )
    DB.session.commit()

    # verify we have 1 record
    assert crud.count_users() == 1
    updated_user = crud.last_user()
    # assert identifier remained the same
    assert updated_user.identifier != new_email
    # assert changes
    assert updated_user.email == new_email
    assert updated_user.affiliation == new_affiliation
    assert updated_user.hashed_password != old_hashed_password
    assert updated_user.public == new_public


# verify reset password
def test_update_password(setup_teardown):
    user = crud.create_user(DB)
    old_hashed_password = user.hashed_password

    new_password = "NewPassword@123"
    user.reset_password(new_password)
    DB.session.commit()

    # verify we have 1 record
    assert crud.count_users() == 1
    updated_user = crud.last_user()
    assert updated_user.hashed_password != old_hashed_password


# verify setting token and salt
def test_set_token(setup_teardown):
    user = crud.create_user(DB)

    assert user.token is None
    assert user.token_created_at is None

    user.set_token_data("secret", "salt")
    DB.session.commit()

    # verify we have 1 record
    assert crud.count_users() == 1
    updated_user = crud.last_user()

    assert updated_user.token is not None
    assert updated_user.token_created_at is not None
    assert isinstance(updated_user.token_created_at, dt)


# verify token validity, by default token is 24 hours valid
@pytest.mark.parametrize(
    "subtract_time", [(10, 0, True), (23, 59, True), (24, 1, False), (25, 0, False)]
)
def test_token_validity(setup_teardown, subtract_time):
    subtract_hours, subtract_mins, validity = subtract_time
    user = crud.create_user(DB)
    user.set_token_data("secret", "salt")
    DB.session.commit()
    # verify we have 1 record
    assert crud.count_users() == 1

    # assert token is valid
    token = user.token
    token_created_at = user.token_created_at
    assert user.token_valid(token)

    # now subtract hours
    new_token_created_time = token_created_at - timedelta(
        hours=subtract_hours, minutes=subtract_mins
    )
    # update token_created_at
    user.token_created_at = new_token_created_time

    # assert token validity
    assert user.token_valid(token) == validity


# test confirming a user
def test_confirm_user(setup_teardown):
    user = crud.create_user(DB)
    # create a token for good measures
    user.set_token_data("secret", "salt")

    assert user.confirmed is False
    assert bool(user.token)
    assert bool(user.token_created_at)

    # now lets confirm
    user.confirm_user()

    assert user.confirmed
    assert user.token is None
    assert user.token_created_at is None


# #############
# DELETE
# #############


# test deleting a user means deleting all projects
def test_deleting_user(setup_teardown):
    user, projects = crud.create_user1_with_2_projects(DB)
    assert crud.count_users() == 1
    assert crud.count_projects() == 2
    # remove the user
    DB.session.delete(user)
    DB.session.commit()
    assert crud.count_users() == 0
    # projects should be gone as well
    assert crud.count_projects() == 0


# #############
# PROPERTIES
# #############


# test projects
def test_projects_of_user(setup_teardown):
    crud.create_user1_with_2_projects(DB)
    assert crud.count_users() == 1
    assert crud.count_projects() == 2
    # get user
    user = crud.last_user()
    projects = crud.list_projects()
    assert set(user.projects) == set(projects)


# test pending invitations
def test_pending_invitations(setup_teardown):
    user1, _ = crud.create_user1_with_2_projects(DB)
    user2 = crud.create_user(DB, user=2)
    assert crud.count_users() == 2
    assert crud.count_projects() == 2
    user1 = crud.get_user_by_id(user1.id)
    project = user1.projects[0]
    project.pending_invitations.append(user2)
    DB.session.commit()
    # fresh object
    user2 = crud.get_user_by_id(user2.id)
    assert project in user2.pending_invitations


# test collaborations
def test_collaboration(setup_teardown):
    user1, _ = crud.create_user1_with_2_projects(DB)
    user2 = crud.create_user(DB, user=2)
    assert crud.count_users() == 2
    assert crud.count_projects() == 2
    user1 = crud.get_user_by_id(user1.id)
    project = user1.projects[0]
    project.collaborators.append(user2)
    DB.session.commit()
    # fresh object
    user2 = crud.get_user_by_id(user2.id)
    assert project in user2.involved_in



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import sys
from pathlib import Path

from filelock import FileLock
from filelock import Timeout

from asreview.models.balance import get_balance_model
from asreview.models.classifiers import get_classifier
from asreview.models.feature_extraction import get_feature_model
from asreview.models.query import get_query_model
from asreview.project import ASReviewProject
from asreview.project import open_state
from asreview.review.base import BaseReview
from asreview.webapp.io import read_data


def get_lab_reviewer(
    as_data,
    project,
    embedding_fp=None,
    verbose=0,
    prior_idx=None,
    prior_record_id=None,
    seed=None,
    **kwargs,
):
    """Get a review object from arguments."""

    if len(as_data) == 0:
        raise ValueError("Supply at least one dataset" " with at least one record.")

    with open_state(project) as state:
        settings = state.settings

    # TODO: Set random seed.
    # Initialize models.
    # random_state = get_random_state(seed)
    classifier_model = get_classifier(settings.model)
    query_model = get_query_model(settings.query_strategy)
    balance_model = get_balance_model(settings.balance_strategy)
    feature_model = get_feature_model(settings.feature_extraction)

    # LSTM models need embedding matrices.
    if classifier_model.name.startswith("lstm-"):
        classifier_model.embedding_matrix = feature_model.get_embedding_matrix(
            as_data.texts, embedding_fp
        )

    # prior knowledge
    if (
        prior_idx is not None
        and prior_record_id is not None
        and len(prior_idx) > 0
        and len(prior_record_id) > 0
    ):
        raise ValueError("Not possible to provide both prior_idx and prior_record_id")

    reviewer = BaseReview(
        as_data,
        project,
        model=classifier_model,
        query_model=query_model,
        balance_model=balance_model,
        feature_model=feature_model,
        **kwargs,
    )
    return reviewer


def train_model(project):
    """Add the new labels to the review and do the modeling.

    It uses a lock to ensure only one model is running at the same time.
    """

    logging.info(f"Project {project.project_path} - Train a new model for project")

    # Lock so that only one training run is running at the same time.
    lock = FileLock(Path(project.project_path, "training.lock"), timeout=0)

    with lock:

        # Check if there are new labeled records.
        with open_state(project.project_path) as state:
            exist_new_labeled_records = state.exist_new_labeled_records

        if exist_new_labeled_records:
            # collect command line arguments and pass them to the reviewer
            as_data = read_data(project)

            reviewer = get_lab_reviewer(as_data, project)

            # Train the model.
            reviewer.train()


def main(argv):
    # parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("project_path", type=str, help="Project id")
    parser.add_argument(
        "--output_error",
        dest="output_error",
        action="store_true",
        help="Save training error message to file.",
    )
    parser.add_argument(
        "--first_run",
        dest="first_run",
        action="store_true",
        help="After first run, status is updated.",
    )
    args = parser.parse_args(argv)

    project = ASReviewProject(args.project_path)

    try:
        train_model(project)

        # change the project status to review
        project.update_review(status="review")

    except Timeout:
        logging.debug("Another iteration is training")

    except Exception as err:
        # save the error to the project
        project.set_error(err, save_error_message=args.output_error)

        # raise the error for full traceback
        raise err
    else:
        project.update_review(status="review")


if __name__ == "__main__":
    main(sys.argv)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class TSVWriter:
    """TSV file writer."""

    name = "tsv"
    label = "TSV (UTF-8)"
    write_format = ".tsv"

    @classmethod
    def write_data(cls, df, fp, sep="\t", labels=None, ranking=None):
        """Export dataset.

        Arguments
        ---------
        df: pandas.Dataframe
            Dataframe of all available record data.
        fp: str, NoneType
            Filepath or None for buffer.
        sep: str
            Seperator of the file.
        labels: list, numpy.ndarray
            Current labels will be overwritten by these labels
            (including unlabelled). No effect if labels is None.
        ranking: list
            Reorder the dataframe according to these (internal) indices.
            Default ordering if ranking is None.

        Returns
        -------
        TSV file
            Dataframe of all available record data.
        """
        return df.to_csv(fp, sep=sep, index=True)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.entry_points.base import BaseEntryPoint
from asreview.webapp.run_model import main as main_run_model
from asreview.webapp.start_flask import main as main_flask


class LABEntryPoint(BaseEntryPoint):
    """Entry point to start the ASReview LAB webapp."""

    def execute(self, argv):

        main_flask(argv)


class WebRunModelEntryPoint(BaseEntryPoint):
    description = "Internal use only."

    def execute(self, argv):
        main_run_model(argv)
